import freetype

enums = [
    'ft_fstypes',
    'ft_face_flags',
    'ft_encodings',
    'ft_glyph_bbox_modes',
    'ft_glyph_formats',
    'ft_kerning_modes',
    'ft_lcd_filters',
    'ft_load_flags',
    'ft_load_targets',
    'ft_open_modes',
    'ft_outline_flags',
    'ft_pixel_modes',
    'ft_render_modes',
    'ft_stroker_borders',
    'ft_stroker_linecaps',
    'ft_stroker_linejoins',
    'ft_style_flags',
    'tt_adobe_ids',
    'tt_apple_ids',
    'tt_mac_ids',
    'tt_ms_ids',
    'tt_ms_langids',
    'tt_mac_langids',
    'tt_name_ids',
    'tt_platforms'
]

for name in enums:
    print name
    module = getattr(freetype, name)
    doc = getattr(module, '__doc__')
    doc = doc.split('\n')
    file = open( name+'.rst', 'w')

    title = name.upper()
    file.write(title+'\n')
    file.write('='*len(title)+'\n')

    for line in doc:
        if line.startswith('FT_') or line.startswith('TT_'):
            file.write( '.. data:: '+ line + '\n')
        else:
            file.write( line + '\n')
    file.close()




# skia_ot_svg_module
# Copyright 2023 Hin-Tak Leung

# Typical usage:
#
#     from freetype import get_handle, FT_Property_Set
#     from skia_ot_svg_module import hooks
#     ...
#     library = get_handle()
#     FT_Property_Set( library, b"ot-svg", b"svg-hooks", byref(hooks) )
#
# After these lines (or equivalent), COLOR OT-SVG font support is enabled.
# You just use freetype-py as usual. Color bitmaps are returned for OT-SVG
# fonts in FT_Load_Char()/FT_Load_Glyph().

from freetype import FT_SVG_Document, FT_Err_Ok, SVG_RendererHooks, \
    SVG_Lib_Init_Func, SVG_Lib_Free_Func, SVG_Lib_Render_Func, \
    SVG_Lib_Preset_Slot_Func, SizeMetrics, FT_UShort, FT_PIXEL_MODE_BGRA, \
    FT_GLYPH_FORMAT_BITMAP
import ctypes
from ctypes import py_object, pythonapi, cast, c_char_p

import skia
from skia import Size, ImageInfo, ColorType, AlphaType, Canvas, \
    PictureRecorder, Rect, ScalarNegativeInfinity, ScalarInfinity, RTreeFactory

from packaging import version
assert(version.parse(skia.__version__) > version.parse("116.0b2")), "Needs Skia-Python 116.0b2+"

pythonapi.PyMemoryView_FromMemory.restype = py_object

# include/private/base/SkFixed.h
def SkFixedToFloat(x):
    return ((x) * 1.52587890625e-5)

from math import ceil, floor

_state = None

def svg_init(ctx):
    global _state
    _state = {}
    ctx.contents.value = _state
    return FT_Err_Ok

def svg_free(ctx):
    global _state
    _state = None
    # "None" is strictly speaking a special pyobject,
    # this line does not do what it should, i.e. setting the
    # pointer to NULL.
    ctx.contents = None
    return # void

def svg_render(slot, ctx):
    state = ctx.contents.value
    dstBitmap = skia.Bitmap()
    dstBitmap.setInfo(ImageInfo.Make(slot.contents.bitmap.width, slot.contents.bitmap.rows,
                                     ColorType.kBGRA_8888_ColorType,
                                     AlphaType.kPremul_AlphaType),
                      slot.contents.bitmap.pitch)
    dstBitmap.setPixels(pythonapi.PyMemoryView_FromMemory(cast(slot.contents.bitmap.buffer, c_char_p),
                                                          slot.contents.bitmap.rows * slot.contents.bitmap.pitch,
                                                          0x200), # Read-Write
                        )
    canvas = Canvas(dstBitmap)

    canvas.clear(skia.ColorTRANSPARENT)

    canvas.translate( -state['x'], -state['y'] )
    canvas.drawPicture( state['picture'] )

    slot.contents.bitmap.pixel_mode = FT_PIXEL_MODE_BGRA
    slot.contents.bitmap.num_grays  = 256
    slot.contents.format            = FT_GLYPH_FORMAT_BITMAP

    state['picture'] = None

    return FT_Err_Ok

def svg_preset_slot(slot, cached, ctx):
    state = ctx.contents.value

    document = ctypes.cast(slot.contents.other, FT_SVG_Document)

    metrics        = SizeMetrics(document.contents.metrics)

    units_per_EM   = FT_UShort(document.contents.units_per_EM)
    end_glyph_id   = FT_UShort(document.contents.end_glyph_id)
    start_glyph_id = FT_UShort(document.contents.start_glyph_id)

    doc = ctypes.string_at(document.contents.svg_document, # not terminated
                           size=document.contents.svg_document_length)
    data = skia.Data(doc)
    svgmem = skia.MemoryStream(data)
    svg = skia.SVGDOM.MakeFromStream(svgmem)

    if (svg.containerSize().isEmpty()):
        size = Size.Make(units_per_EM.value, units_per_EM.value)
        svg.setContainerSize(size)

    recorder = PictureRecorder()

    infiniteRect = Rect.MakeLTRB(ScalarNegativeInfinity, ScalarNegativeInfinity,
                                 ScalarInfinity, ScalarInfinity)
    bboxh = RTreeFactory()()

    recordingCanvas = recorder.beginRecording(infiniteRect, bboxh)

    ftMatrix = document.contents.transform
    ftOffset = document.contents.delta

    m = skia.Matrix()
    m.setAll(
        SkFixedToFloat(ftMatrix.xx), -SkFixedToFloat(ftMatrix.xy),  SkFixedToFloat(ftOffset.x),
       -SkFixedToFloat(ftMatrix.yx),  SkFixedToFloat(ftMatrix.yy), -SkFixedToFloat(ftOffset.y),
        0                          ,  0                          ,  1                        )

    m.postScale(SkFixedToFloat(document.contents.metrics.x_scale) / 64.0,
                SkFixedToFloat(document.contents.metrics.y_scale) / 64.0)

    recordingCanvas.concat(m)

    if ( start_glyph_id.value < end_glyph_id.value ):
        id = "glyph%u" % ( slot.contents.glyph_index )
        svg.renderNode(recordingCanvas, id)
    else:
        svg.render(recordingCanvas)

    state['picture'] = recorder.finishRecordingAsPicture()
    bounds = state['picture'].cullRect()

    width  = ceil(bounds.right()) - floor(bounds.left())
    height = ceil(bounds.bottom()) - floor(bounds.top())
    x = floor(bounds.left())
    y = floor(bounds.top())

    state['x'] = x
    state['y'] = y

    slot.contents.bitmap_left = int(state['x']) # float to int conversion
    slot.contents.bitmap_top  = int(-state['y'])

    slot.contents.bitmap.rows  = ceil( height ) # float to int
    slot.contents.bitmap.width = ceil( width )

    slot.contents.bitmap.pitch = slot.contents.bitmap.width * 4

    slot.contents.bitmap.pixel_mode = FT_PIXEL_MODE_BGRA

    metrics_width  = width
    metrics_height = height

    horiBearingX =  state['x']
    horiBearingY = -state['y']

    vertBearingX = slot.contents.metrics.horiBearingX / 64.0 - slot.contents.metrics.horiAdvance / 64.0 / 2
    vertBearingY = ( slot.contents.metrics.vertAdvance / 64.0 - slot.contents.metrics.height / 64.0 ) / 2

    slot.contents.metrics.width  = int(round( width * 64 ))
    slot.contents.metrics.height = int(round( height * 64 ))

    slot.contents.metrics.horiBearingX = int( horiBearingX * 64 )
    slot.contents.metrics.horiBearingY = int( horiBearingY * 64 )
    slot.contents.metrics.vertBearingX = int( vertBearingX * 64 )
    slot.contents.metrics.vertBearingY = int( vertBearingY * 64 )

    if ( slot.contents.metrics.vertAdvance == 0 ):
        slot.contents.metrics.vertAdvance = int( height * 1.2 * 64 )

    if ( cached == False ):
        state['picture'] = None
        state['x'] = 0
        state['y'] = 0

    return FT_Err_Ok

hooks = SVG_RendererHooks(svg_init=SVG_Lib_Init_Func(svg_init),
                          svg_free=SVG_Lib_Free_Func(svg_free),
                          svg_render=SVG_Lib_Render_Func(svg_render),
                          svg_preset_slot=SVG_Lib_Preset_Slot_Func(svg_preset_slot))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  pycairo/cairocffi-based glyph-outline example - Copyright 2017 Hin-Tak Leung
#  Distributed under the terms of the new BSD license.
#
#  rewrite of the numply,matplotlib-based example from Nicolas P. Rougier
#
# -----------------------------------------------------------------------------
'''
Glyph outline rendering
'''
from freetype import *

# using Matrix class from Cairo, instead of FreeType's
from cairo import Context, ImageSurface, FORMAT_ARGB32, SurfacePattern, FILTER_BEST, Matrix
from bitmap_to_surface import make_image_surface

if __name__ == '__main__':
    from PIL import Image

    face = Face('./Vera.ttf')
    face.set_char_size( 4*48*64 )
    flags = FT_LOAD_DEFAULT | FT_LOAD_NO_BITMAP
    face.load_char('S', flags )
    slot = face.glyph
    glyph = slot.get_glyph()
    stroker = Stroker( )
    stroker.set(64, FT_STROKER_LINECAP_ROUND, FT_STROKER_LINEJOIN_ROUND, 0 )
    glyph.stroke( stroker , True )
    blyph = glyph.to_bitmap(FT_RENDER_MODE_NORMAL, Vector(0,0), True )
    bitmap = blyph.bitmap
    width, rows, pitch = bitmap.width, bitmap.rows, bitmap.pitch
    surface = ImageSurface(FORMAT_ARGB32, 600, 800)
    ctx = Context(surface)
    Z = make_image_surface(bitmap)
    ctx.set_source_surface(Z, 0, 0)
    scale = 640.0 / rows
    patternZ = ctx.get_source()
    SurfacePattern.set_filter(patternZ, FILTER_BEST)
    scalematrix = Matrix()
    scalematrix.scale(1.0/scale,1.0/scale)
    scalematrix.translate(-(300.0 - width *scale /2.0 ), -80)
    patternZ.set_matrix(scalematrix)
    ctx.set_source_rgb (0 , 0, 1)
    ctx.mask(patternZ)
    ctx.fill()
    surface.flush()
    surface.write_to_png("glyph-outline-cairo.png")
    Image.open("glyph-outline-cairo.png").show()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
"""
These values determine how two joining lines are rendered in a stroker.


FT_STROKER_LINEJOIN_ROUND

  Used to render rounded line joins. Circular arcs are used to join two lines
  smoothly.


FT_STROKER_LINEJOIN_BEVEL

  Used to render beveled line joins; i.e., the two joining lines are extended
  until they intersect.


FT_STROKER_LINEJOIN_MITER

  Same as beveled rendering, except that an additional line break is added if
  the angle between the two joining lines is too closed (this is useful to
  avoid unpleasant spikes in beveled rendering).
"""
FT_STROKER_LINEJOINS = { 'FT_STROKER_LINEJOIN_ROUND' : 0,
                         'FT_STROKER_LINEJOIN_BEVEL' : 1,
                         'FT_STROKER_LINEJOIN_MITER' : 2}
globals().update(FT_STROKER_LINEJOINS)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
'''
Show glyph metrics in horizontal and vertical layout
'''
from freetype import *

def arrow( x,y, dx, dy, **kwargs):
    kwargs['shape'] = 'full'
    kwargs['head_width'] = 30
    kwargs['head_length'] = 40
    kwargs['length_includes_head'] =True
    kwargs['facecolor'] = 'k'
    kwargs['edgecolor'] ='k'
    kwargs['linewidth'] =.5
    plt.arrow(x,y,dx,dy,**kwargs)

def double_arrow(x, y, dx, dy, **kwargs):
    cx,cy = x+dx/2., y+dy/2.
    dx /= 2.0
    dy /= 2.0
    arrow(cx,cy,+dx,+dy,**kwargs)
    arrow(cx,cy,-dx,-dy,**kwargs)

def line(x, y, dx, dy, **kwargs):
    kwargs['color'] = 'k'
    kwargs['linewidth'] =.5
    plt.plot([x,x+dx],[y,y+dy],**kwargs)

def point(x, y, r, **kwargs):
    kwargs['color'] = 'k'
    plt.scatter([x],[y],r,**kwargs)

def text( x,y,text, **kwargs):
    kwargs['fontsize'] = 18
    plt.text(x, y, text, **kwargs)



if __name__ == '__main__':
    import numpy as np
    import matplotlib.pyplot as plt
    from matplotlib.path import Path
    import matplotlib.patches as patches


    face = Face('./Vera.ttf')
    face.set_char_size( 32*64 )
    face.load_char('g')
    slot = face.glyph
    bitmap = slot.bitmap
    width  = slot.bitmap.width
    rows   = slot.bitmap.rows
    pitch  = slot.bitmap.pitch
    outline= slot.outline

    start, end = 0, 0
    VERTS, CODES = [], []
    # Iterate over each contour
    for i in range(len(outline.contours)):
        end    = outline.contours[i]
        points = outline.points[start:end+1]
        points.append(points[0])
        tags   = outline.tags[start:end+1]
        tags.append(tags[0])
        segments = [ [points[0],], ]
        for j in range(1, len(points) ):
            segments[-1].append(points[j])
            if tags[j] & (1 << 0) and j < (len(points)-1):
                segments.append( [points[j],] )
        verts = [points[0], ]
        codes = [Path.MOVETO,]
        for segment in segments:
            if len(segment) == 2:
                verts.extend(segment[1:])
                codes.extend([Path.LINETO])
            elif len(segment) == 3:
                verts.extend(segment[1:])
                codes.extend([Path.CURVE3, Path.CURVE3])
            else:
                verts.append(segment[1])
                codes.append(Path.CURVE3)
                for i in range(1,len(segment)-2):
                    A,B = segment[i], segment[i+1]
                    C = ((A[0]+B[0])/2.0, (A[1]+B[1])/2.0)
                    verts.extend([ C, B ])
                    codes.extend([ Path.CURVE3, Path.CURVE3])
                verts.append(segment[-1])
                codes.append(Path.CURVE3)
        VERTS.extend(verts)
        CODES.extend(codes)
        start = end+1
    VERTS = np.array(VERTS)
    x,y = VERTS[:,0], VERTS[:,1]
    VERTS[:,0], VERTS[:,1] = x, y


    path = Path(VERTS, CODES)
    xmin, xmax = x.min(), x.max()
    ymin, ymax = y.min(), y.max()
    width,height = xmax-xmin, ymax-ymin
    dw, dh = 0.2*width, 0.1*height
    bearing = xmin - slot.metrics.horiBearingX, ymin - slot.metrics.horiBearingY
    advance = slot.advance
    origin  = bearing


    figure = plt.figure(figsize=(16,10), frameon=False, facecolor="white")

    axes = plt.subplot(121, frameon=False, aspect=1)
    glyph = patches.PathPatch(path, fill = True, facecolor='k', lw=0)
    plt.xlim(xmin - .25*width,  xmax + .75*width)
    plt.ylim(ymin - .5*height, xmax + .75*height)
    plt.xticks([]), plt.yticks([])
    axes.add_patch(glyph)

    # Y axis
    arrow(origin[0], ymin-dh, 0, height+3*dh)

    # X axis
    arrow(origin[0]-dw, 0, width+3*dw, 0)

    # origin
    point(0,0,50)
    text( -20, -20, "$origin$", va='top', ha='right')

    # Bounding box
    bbox = patches.Rectangle( (xmin,ymin), width, height, fill = False, lw=.5)
    axes.add_patch(bbox)

    # Width
    line(xmin, ymax, 0, 3*dh, linestyle="dotted")
    text( xmin, ymax+3.25*dh, "$x_{min}$", va='bottom', ha='center')
    line(xmax, ymax, 0, 3*dh, linestyle="dotted")
    text( xmax, ymax+3.25*dh, "$x_{max}$", va='bottom', ha='center')
    double_arrow(xmin, ymax+2.5*dh, width, 0)
    text(xmin+width/2., ymax+1.75*dh, "$width$", va='bottom', ha='center')

    # Height
    line(xmax, ymin, 3*dw, 0, linestyle="dotted")
    text(xmax+3.25*dw, ymin, "$y_{min}$", va='baseline', ha='left')
    line(xmax, ymax, 3*dw, 0, linestyle="dotted")
    text(xmax+3.25*dw, ymax, "$y_{max}$", va='baseline', ha='left')
    double_arrow(xmax+2.5*dw, ymin, 0, height)
    text(xmax+2.75*dw, ymin+height/2., "$height$", va='center', ha='left')

    # Advance
    point(advance.x,0,50)
    line(advance.x, 0, 0, ymin-dh, linestyle="dotted")
    arrow(0, ymin-.5*dh, advance.x, 0)
    text(advance.x/2., ymin-1.25*dh, "$advance$", va='bottom', ha='center')

    # Bearing Y
    arrow(xmax+.25*dw, 0, 0, ymax)
    text(xmax+.5*dw, ymax/2, "$Y_{bearing}$", va='center', ha='left')

    # Bearing X
    arrow(0, ymax/2., xmin, 0)
    text(-10, ymax/2, "$X_{bearing}$", va='baseline', ha='right')


    # -------------------------------------------------------------------------

    axes = plt.subplot(122, frameon=False, aspect=1)
    glyph = patches.PathPatch(path, fill = True, facecolor='k', lw=0)
    axes.add_patch(glyph)

    plt.xlim(xmin - .25*width,  xmax + .75*width)
    plt.ylim(ymin - .5*height, xmax + .75*height)
    plt.xticks([]), plt.yticks([])


    advance   = slot.metrics.vertAdvance
    x_bearing = slot.metrics.vertBearingX
    y_bearing = slot.metrics.vertBearingY

    # Y axis
    arrow(xmin-x_bearing, ymax+y_bearing+2*dh, 0, -advance-3*dh)

    # X axis
    arrow(xmin-2*dw, ymax+y_bearing, width+4*dw, 0)

    # origin
    point( xmin-x_bearing, ymax+y_bearing, 50)
    text( xmin-x_bearing-30, ymax+y_bearing+10, "$origin$", va='bottom', ha='right')

    # Bounding box
    bbox = patches.Rectangle( (xmin,ymin), width, height, fill = False, lw=.5)
    axes.add_patch(bbox)


    # # Advance
    point(xmin-x_bearing, ymax+y_bearing-advance, 50)
    line(xmin-x_bearing, ymax+y_bearing-advance, xmax-dw, 0, linestyle="dotted")
    arrow(xmax+dw, ymax+y_bearing, 0, -advance)
    text(xmax+1.25*dw, ymax+y_bearing-advance/2., "$advance$", va='baseline', ha='left')


    # Width
    line(xmin, ymin, 0, -4*dh, linestyle="dotted")
    text( xmin, ymin-4.25*dh, "$x_{min}$", va='top', ha='center')
    line(xmax, ymin, 0, -4*dh, linestyle="dotted")
    text( xmax, ymin-4.25*dh, "$x_{max}$", va='top', ha='center')
    double_arrow(xmin, ymin-3.5*dh, width, 0)
    text(xmin+width/2., ymin-3.75*dh, "$width$", va='top', ha='center')

    # Height
    line(xmin, ymin, -3*dw, 0, linestyle="dotted")
    text(xmin-1.5*dw, ymin, "$y_{min}$", va='baseline', ha='right')
    line(xmin, ymax, -3*dw, 0, linestyle="dotted")
    text(xmin-1.5*dw, ymax, "$y_{max}$", va='baseline', ha='right')
    double_arrow(xmin-.5*dw, ymin, 0, height)
    text(xmin-.75*dw, ymin+height/2., "$height$", va='center', ha='right')


    #point(xmin-x_bearing, ymax+y_bearing, 50)
    # Bearing Y
    arrow(xmax-.5*dw, ymax+y_bearing, 0, -y_bearing)
    text(xmax-.5*dw, ymax+y_bearing+.25*dh, "$Y_{bearing}$", va='bottom', ha='center')

    # # Bearing X
    line(xmin, ymax, 0, 3*dh, linestyle="dotted")
    arrow(xmin-x_bearing, ymax+y_bearing+dh, x_bearing, 0)
    text(xmin-.25*dw,  ymax+y_bearing+dh, "$X_{bearing}$", va='baseline', ha='right')

    plt.savefig('glyph-metrics.pdf')
    plt.show()



def test_pyi_freetype(pyi_builder):
    pyi_builder.test_source(
        """
        import sys
        import pathlib

        import freetype

        # Ensure that the freetype shared library is bundled with the frozen
        # application; otherwise, freetype might be using system-wide library.

        # Check that freetype.FT_Library_filename is an absolute path;
        # otherwise, it is likely using basename-only ctypes fallback.
        ft_library_file = pathlib.Path(freetype.FT_Library_filename)
        print(f"FT library file (original): {ft_library_file}", file=sys.stderr)
        assert ft_library_file.is_absolute(), \
            "FT library file is not an absolute path!"

        # Check that fully-resolved freetype.FT_Library_filename is
        # anchored in fully-resolved frozen application directory.
        app_dir = pathlib.Path(__file__).resolve().parent
        print(f"Application directory: {app_dir}", file=sys.stderr)

        ft_library_path = pathlib.Path(ft_library_file).resolve()
        print(f"FT library file (resolved): {ft_library_path}", file=sys.stderr)

        assert app_dir in ft_library_path.parents, \
            "FT library is not bundled with frozen application!"
        """
    )



import os

import freetype
import pytest

test_folder = os.path.realpath(os.path.dirname(__file__))

def _ft_ver():
    ft_ver_tuple = freetype.version()
    ft_ver_str = ".".join(['{}'.format(i) for i in ft_ver_tuple])

    return ft_ver_tuple, ft_ver_str


def test_variable_font_basics():
    """
    Check that basic VF functionality functions (open a VF font, get VF info)
    """
    ft_vt, ft_vs = _ft_ver()
    if ft_vt < (2, 8, 1):
        # note: there is some proto-VF (Multiple Master) support in older
        # FreeType versions, but not enough to work with here.
        pytest.skip("Incomplete VF support in FreeType lib {} (need 2.8.1 or later)".format(ft_vs))
        return

    font_path = os.path.join(
        test_folder,
        "..",
        "examples",
        "SourceSansVariable-Roman.otf")
    face = freetype.Face(font_path)
    var_info = face.get_variation_info()

    assert len(var_info.axes) == 1
    assert var_info.axes[0].tag == 'wght'

    assert len(var_info.instances) == 6
    for ii, cv in enumerate((200, 300, 400, 600, 700, 900)):
        assert var_info.instances[ii].coords == (cv,)


def test_vf_axis_get_set():
    ft_vt, ft_vs = _ft_ver()
    if ft_vt < (2, 8, 1):
        pytest.skip("Incomplete VF support in FreeType lib {} (need 2.8.1 or later)".format(ft_vs))
        return

    font_path = os.path.join(
        test_folder,
        "..",
        "examples",
        "SourceSansVariable-Roman.otf")

    face = freetype.Face(font_path)

    dc_in = (373,)
    face.set_var_design_coords(dc_in)
    dcoords = face.get_var_design_coords()
    assert dcoords[0] == dc_in[0]

    bc_in = (-0.5,)
    face.set_var_blend_coords(bc_in)
    bcoords = face.get_var_blend_coords()
    assert bcoords[0] == bc_in[0]

    # try 'reset to defaults' option
    face.set_var_design_coords(None, reset=True)
    dcoords = face.get_var_design_coords()
    vsi = face.get_variation_info()
    expected = tuple([vsi.axes[i].default for i in range(len(dcoords))])
    assert dcoords == expected


def test_vf_set_named_instance():
    ft_vt, ft_vs = _ft_ver()
    if ft_vt < (2, 8, 1):
        pytest.skip("Incomplete VF support in FreeType lib {} (need 2.8.1 or later)".format(ft_vs))
        return

    font_path = os.path.join(
        test_folder,
        "..",
        "examples",
        "SourceSansVariable-Roman.otf")

    face = freetype.Face(font_path)
    vsi = face.get_variation_info()

    for inst in vsi.instances:
        face.set_var_named_instance(inst.name)
        dcoords = face.get_var_design_coords()
        assert dcoords == inst.coords



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2018 Dean Serenevy <dean@serenevy.net>
# SPDX-License-Identifier: BSD-3-Clause-Clear
'''
Show how to use outline decompose.
'''
import os
import sys

sys.path.append(os.path.dirname(__file__))
sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

import freetype

def move_to(a, ctx):
    ctx.append("M {},{}".format(a.x, a.y))

def line_to(a, ctx):
    ctx.append("L {},{}".format(a.x, a.y))

def conic_to(a, b, ctx):
    ctx.append("Q {},{} {},{}".format(a.x, a.y, b.x, b.y))

def cubic_to(a, b, c, ctx):
    ctx.append("C {},{} {},{} {},{}".format(a.x, a.y, b.x, b.y, c.x, c.y))

if __name__ == '__main__':
    face = freetype.Face('./Vera.ttf')
    face.set_char_size( 18*64 )
    face.load_char('B', freetype.FT_LOAD_DEFAULT | freetype.FT_LOAD_NO_BITMAP)
    ctx = []
    face.glyph.outline.decompose(ctx, move_to=move_to, line_to=line_to, conic_to=conic_to, cubic_to=cubic_to)
    print("""<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<svg xmlns="http://www.w3.org/2000/svg">
  <g>
    <path
      style="fill:none;stroke:#000000;stroke-width:2;stroke-linecap:butt;stroke-linejoin:miter;stroke-miterlimit:4;stroke-opacity:1;stroke-dasharray:none;stroke-dashoffset:0"
      d="{}"
    />
  </g>
</svg>
""".format(" ".join(ctx)))



FT_Color_Root_Transform = {
    'FT_COLOR_INCLUDE_ROOT_TRANSFORM' : 0,
    'FT_COLOR_NO_ROOT_TRANSFORM'      : 1,
    'FT_COLOR_ROOT_TRANSFORM_MAX'     : 2,
}
globals().update(FT_Color_Root_Transform)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
'''
Texture font class

'''
import sys
import math
import numpy as np
import OpenGL.GL as gl
from freetype import *


class TextureAtlas:
    '''
    Group multiple small data regions into a larger texture.

    The algorithm is based on the article by Jukka Jylänki : "A Thousand Ways
    to Pack the Bin - A Practical Approach to Two-Dimensional Rectangle Bin
    Packing", February 27, 2010. More precisely, this is an implementation of
    the Skyline Bottom-Left algorithm based on C++ sources provided by Jukka
    Jylänki at: http://clb.demon.fi/files/RectangleBinPack/

    Example usage:
    --------------

    atlas = TextureAtlas(512,512,3)
    region = atlas.get_region(20,20)
    ...
    atlas.set_region(region, data)
    '''

    def __init__(self, width=1024, height=1024, depth=1):
        '''
        Initialize a new atlas of given size.

        Parameters
        ----------

        width : int
            Width of the underlying texture

        height : int
            Height of the underlying texture

        depth : 1 or 3
            Depth of the underlying texture
        '''
        self.width  = int(math.pow(2, int(math.log(width, 2) + 0.5)))
        self.height = int(math.pow(2, int(math.log(height, 2) + 0.5)))
        self.depth  = depth
        self.nodes  = [ (0,0,self.width), ]
        self.data   = np.zeros((self.height, self.width, self.depth),
                               dtype=np.ubyte)
        self.texid  = 0
        self.used   = 0



    def upload(self):
        '''
        Upload atlas data into video memory.
        '''

        if not self.texid:
            self.texid = gl.glGenTextures(1)

        gl.glBindTexture( gl.GL_TEXTURE_2D, self.texid )
        gl.glTexParameteri( gl.GL_TEXTURE_2D,
                            gl.GL_TEXTURE_WRAP_S, gl.GL_CLAMP )
        gl.glTexParameteri( gl.GL_TEXTURE_2D,
                            gl.GL_TEXTURE_WRAP_T, gl.GL_CLAMP )
        gl.glTexParameteri( gl.GL_TEXTURE_2D,
                            gl.GL_TEXTURE_MAG_FILTER, gl.GL_LINEAR )
        gl.glTexParameteri( gl.GL_TEXTURE_2D,
                            gl.GL_TEXTURE_MIN_FILTER, gl.GL_LINEAR )
        if self.depth == 1:
            gl.glTexImage2D( gl.GL_TEXTURE_2D, 0, gl.GL_ALPHA,
                             self.width, self.height, 0,
                             gl.GL_ALPHA, gl.GL_UNSIGNED_BYTE, self.data )
        else:
            gl.glTexImage2D( gl.GL_TEXTURE_2D, 0, gl.GL_RGB,
                             self.width, self.height, 0,
                             gl.GL_RGB, gl.GL_UNSIGNED_BYTE, self.data )



    def set_region(self, region, data):
        '''
        Set a given region width provided data.

        Parameters
        ----------

        region : (int,int,int,int)
            an allocated region (x,y,width,height)

        data : numpy array
            data to be copied into given region
        '''

        x, y, width, height = region
        self.data[y:y+height,x:x+width, :] = data



    def get_region(self, width, height):
        '''
        Get a free region of given size and allocate it

        Parameters
        ----------

        width : int
            Width of region to allocate

        height : int
            Height of region to allocate

        Return
        ------
            A newly allocated region as (x,y,width,height) or (-1,-1,0,0)
        '''

        best_height = sys.maxsize
        best_index = -1
        best_width = sys.maxsize
        region = 0, 0, width, height

        for i in range(len(self.nodes)):
            y = self.fit(i, width, height)
            if y >= 0:
                node = self.nodes[i]
                if (y+height < best_height or
                    (y+height == best_height and node[2] < best_width)):
                    best_height = y+height
                    best_index = i
                    best_width = node[2]
                    region = node[0], y, width, height

        if best_index == -1:
            return -1,-1,0,0

        node = region[0], region[1]+height, width
        self.nodes.insert(best_index, node)

        i = best_index+1
        while i < len(self.nodes):
            node = self.nodes[i]
            prev_node = self.nodes[i-1]
            if node[0] < prev_node[0]+prev_node[2]:
                shrink = prev_node[0]+prev_node[2] - node[0]
                x,y,w = self.nodes[i]
                self.nodes[i] = x+shrink, y, w-shrink
                if self.nodes[i][2] <= 0:
                    del self.nodes[i]
                    i -= 1
                else:
                    break
            else:
                break
            i += 1

        self.merge()
        self.used += width*height
        return region



    def fit(self, index, width, height):
        '''
        Test if region (width,height) fit into self.nodes[index]

        Parameters
        ----------

        index : int
            Index of the internal node to be tested

        width : int
            Width or the region to be tested

        height : int
            Height or the region to be tested

        '''

        node = self.nodes[index]
        x,y = node[0], node[1]
        width_left = width

        if x+width > self.width:
            return -1

        i = index
        while width_left > 0:
            node = self.nodes[i]
            y = max(y, node[1])
            if y+height > self.height:
                return -1
            width_left -= node[2]
            i += 1
        return y



    def merge(self):
        '''
        Merge nodes
        '''

        i = 0
        while i < len(self.nodes)-1:
            node = self.nodes[i]
            next_node = self.nodes[i+1]
            if node[1] == next_node[1]:
                self.nodes[i] = node[0], node[1], node[2]+next_node[2]
                del self.nodes[i+1]
            else:
                i += 1


class TextureFont:
    '''
    A texture font gathers a set of glyph relatively to a given font filename
    and size.
    '''

    def __init__(self, atlas, filename, size):
        '''
        Initialize font

        Parameters:
        -----------

        atlas: TextureAtlas
            Texture atlas where glyph texture will be stored

        filename: str
            Font filename

        size : float
            Font size
        '''
        self.atlas = atlas
        self.filename = filename
        self.size = size
        self.glyphs = {}
        face = Face( self.filename )
        face.set_char_size( int(self.size*64))
        self._dirty = False
        metrics = face.size
        self.ascender  = metrics.ascender/64.0
        self.descender = metrics.descender/64.0
        self.height    = metrics.height/64.0
        self.linegap   = self.height - self.ascender + self.descender
        self.depth = atlas.depth
        try:
            set_lcd_filter(FT_LCD_FILTER_LIGHT)
        except:
            pass


    def __getitem__(self, charcode):
        '''
        x.__getitem__(y) <==> x[y]
        '''
        if charcode not in self.glyphs.keys():
            self.load('%c' % charcode)
        return self.glyphs[charcode]



    def get_texid(self):
        '''
        Get underlying texture identity .
        '''

        if self._dirty:
            self.atlas.upload()
        self._dirty = False
        return self.atlas.texid

    texid = property(get_texid,
                     doc='''Underlying texture identity.''')



    def load(self, charcodes = ''):
        '''
        Build glyphs corresponding to individual characters in charcodes.

        Parameters:
        -----------

        charcodes: [str | unicode]
            Set of characters to be represented
        '''
        face = Face( self.filename )
        pen = Vector(0,0)
        hres = 16*72
        hscale = 1.0/16

        for charcode in charcodes:
            face.set_char_size( int(self.size * 64), 0, hres, 72 )
            matrix = Matrix( int((hscale) * 0x10000), int((0.0) * 0x10000),
                             int((0.0)    * 0x10000), int((1.0) * 0x10000) )
            face.set_transform( matrix, pen )
            if charcode in self.glyphs.keys():
                continue

            self.dirty = True
            flags = FT_LOAD_RENDER | FT_LOAD_FORCE_AUTOHINT
            flags |= FT_LOAD_TARGET_LCD

            face.load_char( charcode, flags )
            bitmap = face.glyph.bitmap
            left   = face.glyph.bitmap_left
            top    = face.glyph.bitmap_top
            width  = face.glyph.bitmap.width
            rows   = face.glyph.bitmap.rows
            pitch  = face.glyph.bitmap.pitch

            x,y,w,h = self.atlas.get_region(width/self.depth+2, rows+2)
            if x < 0:
                print ('Missed !')
                continue
            x,y = x+1, y+1
            w,h = w-2, h-2
            data = []
            for i in range(rows):
                data.extend(bitmap.buffer[i*pitch:i*pitch+width])
            data = np.array(data,dtype=np.ubyte).reshape(h,w,3)
            gamma = 1.5
            Z = ((data/255.0)**(gamma))
            data = (Z*255).astype(np.ubyte)
            self.atlas.set_region((x,y,w,h), data)

            # Build glyph
            size   = w,h
            offset = left, top
            advance= face.glyph.advance.x, face.glyph.advance.y

            u0     = (x +     0.0)/float(self.atlas.width)
            v0     = (y +     0.0)/float(self.atlas.height)
            u1     = (x + w - 0.0)/float(self.atlas.width)
            v1     = (y + h - 0.0)/float(self.atlas.height)
            texcoords = (u0,v0,u1,v1)
            glyph = TextureGlyph(charcode, size, offset, advance, texcoords)
            self.glyphs[charcode] = glyph

            # Generate kerning
            for g in self.glyphs.values():
                # 64 * 64 because of 26.6 encoding AND the transform matrix used
                # in texture_font_load_face (hres = 64)
                kerning = face.get_kerning(g.charcode, charcode, mode=FT_KERNING_UNFITTED)
                if kerning.x != 0:
                    glyph.kerning[g.charcode] = kerning.x/(64.0*64.0)
                kerning = face.get_kerning(charcode, g.charcode, mode=FT_KERNING_UNFITTED)
                if kerning.x != 0:
                    g.kerning[charcode] = kerning.x/(64.0*64.0)

            # High resolution advance.x calculation
            # gindex = face.get_char_index( charcode )
            # a = face.get_advance(gindex, FT_LOAD_RENDER | FT_LOAD_TARGET_LCD)/(64*72)
            # glyph.advance = a, glyph.advance[1]


class TextureGlyph:
    '''
    A texture glyph gathers information relative to the size/offset/advance and
    texture coordinates of a single character. It is generally built
    automatically by a TextureFont.
    '''

    def __init__(self, charcode, size, offset, advance, texcoords):
        '''
        Build a new texture glyph

        Parameter:
        ----------

        charcode : char
            Represented character

        size: tuple of 2 ints
            Glyph size in pixels

        offset: tuple of 2 floats
            Glyph offset relatively to anchor point

        advance: tuple of 2 floats
            Glyph advance

        texcoords: tuple of 4 floats
            Texture coordinates of bottom-left and top-right corner
        '''
        self.charcode = charcode
        self.size = size
        self.offset = offset
        self.advance = advance
        self.texcoords = texcoords
        self.kerning = {}


    def get_kerning(self, charcode):
        ''' Get kerning information

        Parameters:
        -----------

        charcode: char
            Character preceding this glyph
        '''
        if charcode in self.kerning.keys():
            return self.kerning[charcode]
        else:
            return 0



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
'''
Freetype enum types
-------------------

FT_CURVE_TAGS: An enumeration type for each point on an outline to indicate
               whether it describes a point used to control a line segment
               or an arc.

FT_PIXEL_MODES: An enumeration type used to describe the format of pixels in a
                given bitmap. Note that additional formats may be added in the
                future.

FT_GLYPH_BBOX_MODES: The mode how the values of FT_Glyph_Get_CBox are returned.

FT_GLYPH_FORMATS: An enumeration type used to describe the format of a given
                  glyph image. Note that this version of FreeType only supports
                  two image formats, even though future font drivers will be
                  able to register their own format.

FT_ENCODINGS: An enumeration used to specify character sets supported by
              charmaps. Used in the FT_Select_Charmap API function.

FT_RENDER_MODES: An enumeration type that lists the render modes supported by
                 FreeType 2. Each mode corresponds to a specific type of
                 scanline conversion performed on the outline.

FT_LOAD_TARGETS: A list of values that are used to select a specific hinting
                 algorithm to use by the hinter. You should OR one of these
                 values to your 'load_flags' when calling FT_Load_Glyph.

FT_LOAD_FLAGS: A list of bit-field constants used with FT_Load_Glyph to
               indicate what kind of operations to perform during glyph
               loading.

FT_STYLE_FLAGS: A list of bit-flags used to indicate the style of a given
                face. These are used in the 'style_flags' field of FT_FaceRec.

FT_FSTYPES: A list of bit flags that inform client applications of embedding
            and subsetting restrictions associated with a font.

FT_FACE_FLAGS: A list of bit flags used in the 'face_flags' field of the
               FT_FaceRec structure. They inform client applications of
               properties of the corresponding face.

FT_OUTLINE_FLAGS: A list of bit-field constants use for the flags in an
                  outline's 'flags' field.

FT_OPEN_MODES: A list of bit-field constants used within the 'flags' field of
               the FT_Open_Args structure.

FT_KERNING_MODES: An enumeration used to specify which kerning values to return
                  in FT_Get_Kerning.

FT_STROKER_LINEJOINS: These values determine how two joining lines are rendered
                      in a stroker.

FT_STROKER_LINECAPS: These values determine how the end of opened sub-paths are
                     rendered in a stroke.

FT_STROKER_BORDERS: These values are used to select a given stroke border in
                    FT_Stroker_GetBorderCounts and FT_Stroker_ExportBorder.

FT_LCD_FILTERS: A list of values to identify various types of LCD filters.

TT_PLATFORMS: A list of valid values for the 'platform_id' identifier code in
              FT_CharMapRec and FT_SfntName structures.

TT_APPLE_IDS: A list of valid values for the 'encoding_id' for
              TT_PLATFORM_APPLE_UNICODE charmaps and name entries.

TT_MAC_IDS: A list of valid values for the 'encoding_id' for
            TT_PLATFORM_MACINTOSH charmaps and name entries.

TT_MS_IDS: A list of valid values for the 'encoding_id' for
           TT_PLATFORM_MICROSOFT charmaps and name entries.

TT_ADOBE_IDS: A list of valid values for the 'encoding_id' for
              TT_PLATFORM_ADOBE charmaps. This is a FreeType-specific
              extension!

TT_MAC_LANGIDS: Possible values of the language identifier field in the name
                records of the TTF `name' table if the `platform' identifier
                code is TT_PLATFORM_MACINTOSH.

TT_MS_LANGIDS: Possible values of the language identifier field in the name
               records of the TTF `name' table if the `platform' identifier
               code is TT_PLATFORM_MICROSOFT.

TT_NAME_IDS: Possible values of the `name' identifier field in the name
             records of the TTF `name' table.  These values are platform
             independent.
'''
from freetype.ft_enums.ft_color_root_transform import *
from freetype.ft_enums.ft_curve_tags import *
from freetype.ft_enums.ft_fstypes import *
from freetype.ft_enums.ft_face_flags import *
from freetype.ft_enums.ft_encodings import *
from freetype.ft_enums.ft_glyph_bbox_modes import *
from freetype.ft_enums.ft_glyph_formats import *
from freetype.ft_enums.ft_kerning_modes import *
from freetype.ft_enums.ft_lcd_filters import *
from freetype.ft_enums.ft_load_flags import *
from freetype.ft_enums.ft_load_targets import *
from freetype.ft_enums.ft_open_modes import *
from freetype.ft_enums.ft_outline_flags import *
from freetype.ft_enums.ft_pixel_modes import *
from freetype.ft_enums.ft_render_modes import *
from freetype.ft_enums.ft_stroker_borders import *
from freetype.ft_enums.ft_stroker_linecaps import *
from freetype.ft_enums.ft_stroker_linejoins import *
from freetype.ft_enums.ft_style_flags import *
from freetype.ft_enums.tt_adobe_ids import *
from freetype.ft_enums.tt_apple_ids import *
from freetype.ft_enums.tt_mac_ids import *
from freetype.ft_enums.tt_ms_ids import *
from freetype.ft_enums.tt_ms_langids import *
from freetype.ft_enums.tt_mac_langids import *
from freetype.ft_enums.tt_name_ids import *
from freetype.ft_enums.tt_platforms import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# The test file used and known to work is v2.004 of https://github.com/adobe-fonts/source-han-sans/blob/release/OTF/Japanese/SourceHanSans-Regular.otf

import io
import os
import traceback

from typing import Dict

from fontTools.ttLib import TTFont

from ctypes import byref
from freetype import (
    FT_Done_Face,
    FT_Done_FreeType,
    FT_Exception,
    FT_Face,
    FT_Get_First_Char,
    FT_Get_Next_Char,
    FT_Init_FreeType,
    FT_Library,
    FT_New_Memory_Face,
    FT_UInt,
    FT_Get_Char_Index,
    FT_Face_GetVariantSelectors,
    FT_Face_GetCharsOfVariant,
    FT_Face_GetCharVariantIndex,
    FT_Face_GetCharVariantIsDefault
)

_ALL_IVS_NUMBERS  = [x for x in range( 0xe0100, 0xe01f0)] # Supplemental
_ALL_IVS_NUMBERS += [x for x in range( 0xfe00,  0xfe10)]  # IVS

def _read_cmap_uvs(uvsDict):
    global _ALL_IVS_NUMBERS

    all_characters         = {}
    all_default_characters = []
    for item, item_list in uvsDict.items():
        ivs_val = int(item)
        if ivs_val in _ALL_IVS_NUMBERS:
            ivs_chr = chr(ivs_val)
            for character_tuple in item_list:
                character_val = character_tuple[0]
                glyph_name    = character_tuple[1]
                if glyph_name:
                    if glyph_name == '.notdef':
                        continue

                    character     = chr(int(character_val))
                    ivs_character = character + ivs_chr
                    assert ivs_character not in all_characters
                    all_characters[ivs_character] = glyph_name
                else:
                    character     = chr(int(character_val))
                    ivs_character = character + ivs_chr
                    all_default_characters.append(ivs_character)

    return all_characters, all_default_characters

def _read_character(character_value, encoding):  
    if encoding == 'utf_16_be':
        character_string = chr(character_value)  
    else:
        return None

    return character_string

_CMAP_PRIORITY_LIST = [
    (3, 10),  # Windows Unicode full repertoire
    (0,  6),  # Unicode full repertoire (format 13 subtable)
    (0,  4),  # Unicode 2.0 full repertoire
    (3,  1),  # Windows Unicode BMP
    (0,  3),  # Unicode 2.0 BMP
    (0,  2),  # Unicode ISO/IEC 10646
    (0,  1),  # Unicode 1.1
    (0,  0)   # Unicode 1.0
]

def read_fonttools_cmap(font) -> Dict[str, str]:
    global _CMAP_PRIORITY_LIST

    assert isinstance(font, TTFont)
    if not hasattr(font["cmap"], 'tables'):
        return None

    all_characters         = {}
    all_default_characters = []
    best_read_index        = None
    all_tables             = font["cmap"].tables
    for table in all_tables:
        encoding        = table.getEncoding()
        if not encoding:
            continue
        if encoding != 'utf_16_be':
            continue       
        try:
            if table.format == 14:
                if hasattr(table, 'uvsDict'):
                    all_uvs_data, default_characters = _read_cmap_uvs(table.uvsDict)
                    all_default_characters          += default_characters
                    for character, glyphname in all_uvs_data.items():
                        if character in all_characters:
                            assert all_characters[character] == glyphname
                        else:
                            all_characters[character] = glyphname
                else:
                    print('Unknown CMAP Format 14: {}:'.format(vars(table)))

            elif hasattr(table, 'cmap'):
                tuple_value = (table.platformID, table.platEncID)
                if tuple_value in _CMAP_PRIORITY_LIST:
                    index_value = _CMAP_PRIORITY_LIST.index(tuple_value)
                    if best_read_index:
                        if index_value < best_read_index:
                            best_read_index = index_value
                        else:
                            continue
                    else:
                        best_read_index = index_value

                all_items = table.cmap.items()
                length = len(all_items)
                if length == 0:
                    if table.format != 6:
                        print('Unknown CMAP Format {}: {}:'.format(table.format, vars(table)))

                for item in all_items:
                    character = _read_character(item[0], encoding)
                    glyphname = item[1]
                    if glyphname == '.notdef':
                        continue

                    if character is not None:
                        if character in all_characters:
                            if all_characters[character] != glyphname:
                                all_characters[character] = glyphname
                        else:
                            all_characters[character] = glyphname

        except:
            traceback.print_exc()
            continue

    if all_default_characters:
        for ivs_character in all_default_characters:
            first_character = ivs_character[0]
            if first_character in all_characters:
                glyphname                     = all_characters[first_character]
                all_characters[ivs_character] = glyphname

    return all_characters

def read_freetype_cmap(face: FT_Face) -> Dict[str, int]:
    platID     = face.contents.charmap.contents.platform_id
    encodingID = face.contents.charmap.contents.encoding_id
    if platID == 3:
        if encodingID not in [1, 10]:
            return {}

    elif platID == 0: # all unicode
        pass

    else: # everything else
        return {}

    all_characters = []
    gindex         = FT_UInt()
    charcode       = FT_Get_First_Char( face, byref(gindex) )
    while gindex.value != 0:
        character = chr(charcode)
        all_characters.append(character)
        charcode = FT_Get_Next_Char( face, charcode, byref(gindex) )

    variant_selectors_list = FT_Face_GetVariantSelectors(face)
    if bool(variant_selectors_list):
        all_selectors  = []
        selector_value = variant_selectors_list[0]
        index          = 0

        while selector_value != 0:
            all_selectors.append(selector_value)

            index         += 1
            selector_value = variant_selectors_list[index]

        for selector_value in all_selectors:
            character_value_list = FT_Face_GetCharsOfVariant(face, selector_value)
            assert(bool(character_value_list))
            character_value = character_value_list[0]
            index           = 0

            while character_value != 0:
                character = chr(character_value) + chr(selector_value)
                all_characters.append(character)

                index          += 1
                character_value = character_value_list[index]

    character_to_glyphID = {}
    for character in all_characters:
        if len(character) == 2:
            character_value = ord(character[0])
            selector_value  = ord(character[1])
            glyphID         = FT_Face_GetCharVariantIndex(face, character_value, selector_value)
            if glyphID != 0:
                assert character not in character_to_glyphID
                character_to_glyphID[character] = glyphID
        else:
            assert len(character) == 1
            character_value = ord(character)
            glyphID         = FT_Get_Char_Index(face, character_value)
            if glyphID != 0:
                assert character not in character_to_glyphID
                character_to_glyphID[character] = glyphID

    return character_to_glyphID

def _convert_character_to_hex(text: str):
    assert len(text) == 1
    value = ord(text)
    if 0x0000 <= value <= 0xFFFF:
        assert len(hex(value)) <= 6
        return '{0:04x}'.format(value)
    elif value <= 0xFFFFF:
        assert len(hex(value)) <= 7
        return '{0:05x}'.format(value)
    elif value <= 0xFFFFFF:
        assert len(hex(value)) <= 8
        return '{0:06x}'.format(value)
    elif value <= 0xFFFFFFF:
        assert len(hex(value)) <= 9
        return '{0:07x}'.format(value)
    elif value <= 0xFFFFFFFF:
        assert len(hex(value)) <= 9
        return '{0:08x}'.format(value)
    else:
        raise RuntimeError()

def convert_string_to_hex(text: str):
    assert isinstance(text, str)
    result = ''
    for count, character in enumerate(text):
        if count > 0:
            result += '-{}'.format(_convert_character_to_hex(character))
        else:
            result += '{}'.format(_convert_character_to_hex(character))

    return result

if __name__ == "__main__":
    directory   = os.path.dirname(__file__)
    font_path   = os.path.join(directory, 'SourceHanSans-Regular.otf')
    memory_file = io.BytesIO()
    with open(font_path, 'rb') as fontfile:
        memory_file.write(fontfile.read())
        memory_file.seek(0)
    
    fonttools_font = TTFont(memory_file, 0, allowVID=0,
                            ignoreDecompileErrors=True,
                            fontNumber=-1)

    library = FT_Library()
    error   = FT_Init_FreeType(byref(library))
    if error: raise FT_Exception(error)

    freetype_face = FT_Face()
    data          = memory_file.getvalue()
    error         = FT_New_Memory_Face(library, data, len(data), 0, byref(freetype_face))
    if error: raise FT_Exception(error)

    all_freetype_characters  = read_freetype_cmap(freetype_face)
    all_fonttools_characters = read_fonttools_cmap(fonttools_font)

    print('Read {} Free Type Characters'.format(len(all_freetype_characters)))
    print('Read {} Font Tools Characters'.format(len(all_fonttools_characters)))

    print('Checking Mapping')

    for character, glyphID in all_freetype_characters.items():
        glyphname = fonttools_font.getGlyphName(glyphID)
        if character in all_fonttools_characters:
            ft_glyphname = all_fonttools_characters[character]
            if ft_glyphname != glyphname:
                character_hex = convert_string_to_hex(character)
                print('Glyph Mismatch: {} Free Type: {} Font Tools: {}'.format(character_hex, glyphname, ft_glyphname))

        else:
            character_hex = convert_string_to_hex(character)
            print('Glyph Missing in Font Tools: {}'.format(character_hex))

    for character, glyphname in all_fonttools_characters.items():
        if character in all_freetype_characters:
            ft_glyphID   = all_freetype_characters[character]
            ft_glyphname = fonttools_font.getGlyphName(ft_glyphID)
            if ft_glyphname != glyphname:
                character_hex = convert_string_to_hex(character)
                print('Glyph Mismatch: {} Font Tools: {} Free Type: {}'.format(character_hex, glyphname, ft_glyphname))

        else:
            character_hex = convert_string_to_hex(character)
            print('Glyph Missing in Free Type: {}'.format(character_hex))

    print('Finished Checking Mapping')

    FT_Done_Face(freetype_face)
    FT_Done_FreeType(library)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  pycairo/cairocffi-based "Hello World" example - Copyright 2017 Hin-Tak Leung
#  Distributed under the terms of the new BSD license.
#
#  rewrite of the numply,matplotlib-based example from Nicolas P. Rougier
#
# -----------------------------------------------------------------------------
from freetype import *

from cairo import Context, ImageSurface, FORMAT_A8
from bitmap_to_surface import make_image_surface

if __name__ == '__main__':
    from PIL import Image

    face = Face('./Vera.ttf')
    text = 'Hello World !'
    face.set_char_size( 48*64 )
    slot = face.glyph

    # First pass to compute bbox
    width, height, baseline = 0, 0, 0
    previous = 0
    for i,c in enumerate(text):
        face.load_char(c)
        bitmap = slot.bitmap
        height = max(height,
                     bitmap.rows + max(0,-(slot.bitmap_top-bitmap.rows)))
        baseline = max(baseline, max(0,-(slot.bitmap_top-bitmap.rows)))
        kerning = face.get_kerning(previous, c)
        width += (slot.advance.x >> 6) + (kerning.x >> 6)
        previous = c

    Z = ImageSurface(FORMAT_A8, width, height)
    ctx = Context(Z)

    # Second pass for actual rendering
    x, y = 0, 0
    previous = 0
    for c in text:
        face.load_char(c)
        bitmap = slot.bitmap
        top = slot.bitmap_top
        left = slot.bitmap_left
        w,h = bitmap.width, bitmap.rows
        y = height-baseline-top
        kerning = face.get_kerning(previous, c)
        x += (kerning.x >> 6)
        # cairo does not like zero-width bitmap from the space character!
        if (bitmap.width > 0):
            glyph_surface = make_image_surface(face.glyph.bitmap)
            ctx.set_source_surface(glyph_surface, x+left, y)
            ctx.paint()
        x += (slot.advance.x >> 6)
        previous = c
    Z.flush()
    Z.write_to_png("hello-world-cairo.png")
    Image.open("hello-world-cairo.png").show()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
"""
Possible values of the language identifier field in the name records of the
TTF 'name' table if the 'platform' identifier code is TT_PLATFORM_MACINTOSH.

TT_MAC_LANGID_LATIN

TT_MAC_LANGID_MALAY_ARABIC_SCRIPT

TT_MAC_LANGID_HINDI

TT_MAC_LANGID_CATALAN

TT_MAC_LANGID_MARATHI

TT_MAC_LANGID_ICELANDIC

TT_MAC_LANGID_ARABIC

TT_MAC_LANGID_SWAHILI

TT_MAC_LANGID_KHMER

TT_MAC_LANGID_UKRAINIAN

TT_MAC_LANGID_FINNISH

TT_MAC_LANGID_POLISH

TT_MAC_LANGID_NEPALI

TT_MAC_LANGID_UZBEK

TT_MAC_LANGID_TELUGU

TT_MAC_LANGID_MALTESE

TT_MAC_LANGID_AFRIKAANS

TT_MAC_LANGID_CHEWA

TT_MAC_LANGID_BASQUE

TT_MAC_LANGID_CZECH

TT_MAC_LANGID_ROMANIAN

TT_MAC_LANGID_QUECHUA

TT_MAC_LANGID_TAGALOG

TT_MAC_LANGID_HUNGARIAN

TT_MAC_LANGID_AZERBAIJANI_CYRILLIC_SCRIPT

TT_MAC_LANGID_TONGAN

TT_MAC_LANGID_SUNDANESE

TT_MAC_LANGID_JAPANESE

TT_MAC_LANGID_MONGOLIAN

TT_MAC_LANGID_ALBANIAN

TT_MAC_LANGID_NORWEGIAN

TT_MAC_LANGID_SLOVAK

TT_MAC_LANGID_MALAGASY

TT_MAC_LANGID_DZONGKHA

TT_MAC_LANGID_DUTCH

TT_MAC_LANGID_MALAY_ROMAN_SCRIPT

TT_MAC_LANGID_SERBIAN

TT_MAC_LANGID_GERMAN

TT_MAC_LANGID_SOMALI

TT_MAC_LANGID_KOREAN

TT_MAC_LANGID_MONGOLIAN_MONGOLIAN_SCRIPT

TT_MAC_LANGID_CROATIAN

TT_MAC_LANGID_TURKISH

TT_MAC_LANGID_MOLDAVIAN

TT_MAC_LANGID_LAO

TT_MAC_LANGID_ORIYA

TT_MAC_LANGID_BRETON

TT_MAC_LANGID_PASHTO

TT_MAC_LANGID_GUARANI

TT_MAC_LANGID_HEBREW

TT_MAC_LANGID_SLOVENIAN

TT_MAC_LANGID_ESTONIAN

TT_MAC_LANGID_RUNDI

TT_MAC_LANGID_URDU

TT_MAC_LANGID_CHINESE_TRADITIONAL

TT_MAC_LANGID_TATAR

TT_MAC_LANGID_CHINESE_SIMPLIFIED

TT_MAC_LANGID_AZERBAIJANI_ARABIC_SCRIPT

TT_MAC_LANGID_SANSKRIT

TT_MAC_LANGID_KURDISH

TT_MAC_LANGID_FAEROESE

TT_MAC_LANGID_MONGOLIAN_CYRILLIC_SCRIPT

TT_MAC_LANGID_TIGRINYA

TT_MAC_LANGID_THAI

TT_MAC_LANGID_DANISH

TT_MAC_LANGID_KAZAKH

TT_MAC_LANGID_YIDDISH

TT_MAC_LANGID_ESPERANTO

TT_MAC_LANGID_LITHUANIAN

TT_MAC_LANGID_FARSI

TT_MAC_LANGID_LETTISH

TT_MAC_LANGID_VIETNAMESE

TT_MAC_LANGID_PORTUGUESE

TT_MAC_LANGID_IRISH

TT_MAC_LANGID_WELSH

TT_MAC_LANGID_PUNJABI

TT_MAC_LANGID_GREEK

TT_MAC_LANGID_INUKTITUT

TT_MAC_LANGID_FRENCH

TT_MAC_LANGID_GREEK_POLYTONIC

TT_MAC_LANGID_AZERBAIJANI

TT_MAC_LANGID_JAVANESE

TT_MAC_LANGID_SWEDISH

TT_MAC_LANGID_UIGHUR

TT_MAC_LANGID_BENGALI

TT_MAC_LANGID_RUANDA

TT_MAC_LANGID_SINDHI

TT_MAC_LANGID_TIBETAN

TT_MAC_LANGID_ENGLISH

TT_MAC_LANGID_SAAMISK

TT_MAC_LANGID_INDONESIAN

TT_MAC_LANGID_MANX_GAELIC

TT_MAC_LANGID_BYELORUSSIAN

TT_MAC_LANGID_BULGARIAN

TT_MAC_LANGID_GEORGIAN

TT_MAC_LANGID_AZERBAIJANI_ROMAN_SCRIPT

TT_MAC_LANGID_ITALIAN

TT_MAC_LANGID_SCOTTISH_GAELIC

TT_MAC_LANGID_ARMENIAN

TT_MAC_LANGID_GALLA

TT_MAC_LANGID_MACEDONIAN

TT_MAC_LANGID_IRISH_GAELIC

TT_MAC_LANGID_KIRGHIZ

TT_MAC_LANGID_TAMIL

TT_MAC_LANGID_SPANISH

TT_MAC_LANGID_BURMESE

TT_MAC_LANGID_KANNADA

TT_MAC_LANGID_GALICIAN

TT_MAC_LANGID_FLEMISH

TT_MAC_LANGID_TAJIKI

TT_MAC_LANGID_ASSAMESE

TT_MAC_LANGID_SINHALESE

TT_MAC_LANGID_GREELANDIC

TT_MAC_LANGID_AMHARIC

TT_MAC_LANGID_KASHMIRI

TT_MAC_LANGID_AYMARA

TT_MAC_LANGID_GUJARATI

TT_MAC_LANGID_RUSSIAN

TT_MAC_LANGID_TURKMEN

TT_MAC_LANGID_MALAYALAM
"""
TT_MAC_LANGIDS = {
    'TT_MAC_LANGID_ENGLISH'                    :   0,
    'TT_MAC_LANGID_FRENCH'                     :   1,
    'TT_MAC_LANGID_GERMAN'                     :   2,
    'TT_MAC_LANGID_ITALIAN'                    :   3,
    'TT_MAC_LANGID_DUTCH'                      :   4,
    'TT_MAC_LANGID_SWEDISH'                    :   5,
    'TT_MAC_LANGID_SPANISH'                    :   6,
    'TT_MAC_LANGID_DANISH'                     :   7,
    'TT_MAC_LANGID_PORTUGUESE'                 :   8,
    'TT_MAC_LANGID_NORWEGIAN'                  :   9,
    'TT_MAC_LANGID_HEBREW'                     :  10,
    'TT_MAC_LANGID_JAPANESE'                   :  11,
    'TT_MAC_LANGID_ARABIC'                     :  12,
    'TT_MAC_LANGID_FINNISH'                    :  13,
    'TT_MAC_LANGID_GREEK'                      :  14,
    'TT_MAC_LANGID_ICELANDIC'                  :  15,
    'TT_MAC_LANGID_MALTESE'                    :  16,
    'TT_MAC_LANGID_TURKISH'                    :  17,
    'TT_MAC_LANGID_CROATIAN'                   :  18,
    'TT_MAC_LANGID_CHINESE_TRADITIONAL'        :  19,
    'TT_MAC_LANGID_URDU'                       :  20,
    'TT_MAC_LANGID_HINDI'                      :  21,
    'TT_MAC_LANGID_THAI'                       :  22,
    'TT_MAC_LANGID_KOREAN'                     :  23,
    'TT_MAC_LANGID_LITHUANIAN'                 :  24,
    'TT_MAC_LANGID_POLISH'                     :  25,
    'TT_MAC_LANGID_HUNGARIAN'                  :  26,
    'TT_MAC_LANGID_ESTONIAN'                   :  27,
    'TT_MAC_LANGID_LETTISH'                    :  28,
    'TT_MAC_LANGID_SAAMISK'                    :  29,
    'TT_MAC_LANGID_FAEROESE'                   :  30,
    'TT_MAC_LANGID_FARSI'                      :  31,
    'TT_MAC_LANGID_RUSSIAN'                    :  32,
    'TT_MAC_LANGID_CHINESE_SIMPLIFIED'         :  33,
    'TT_MAC_LANGID_FLEMISH'                    :  34,
    'TT_MAC_LANGID_IRISH'                      :  35,
    'TT_MAC_LANGID_ALBANIAN'                   :  36,
    'TT_MAC_LANGID_ROMANIAN'                   :  37,
    'TT_MAC_LANGID_CZECH'                      :  38,
    'TT_MAC_LANGID_SLOVAK'                     :  39,
    'TT_MAC_LANGID_SLOVENIAN'                  :  40,
    'TT_MAC_LANGID_YIDDISH'                    :  41,
    'TT_MAC_LANGID_SERBIAN'                    :  42,
    'TT_MAC_LANGID_MACEDONIAN'                 :  43,
    'TT_MAC_LANGID_BULGARIAN'                  :  44,
    'TT_MAC_LANGID_UKRAINIAN'                  :  45,
    'TT_MAC_LANGID_BYELORUSSIAN'               :  46,
    'TT_MAC_LANGID_UZBEK'                      :  47,
    'TT_MAC_LANGID_KAZAKH'                     :  48,
    'TT_MAC_LANGID_AZERBAIJANI'                :  49,
    'TT_MAC_LANGID_AZERBAIJANI_CYRILLIC_SCRIPT':  49,
    'TT_MAC_LANGID_AZERBAIJANI_ARABIC_SCRIPT'  :  50,
    'TT_MAC_LANGID_ARMENIAN'                   :  51,
    'TT_MAC_LANGID_GEORGIAN'                   :  52,
    'TT_MAC_LANGID_MOLDAVIAN'                  :  53,
    'TT_MAC_LANGID_KIRGHIZ'                    :  54,
    'TT_MAC_LANGID_TAJIKI'                     :  55,
    'TT_MAC_LANGID_TURKMEN'                    :  56,
    'TT_MAC_LANGID_MONGOLIAN'                  :  57,
    'TT_MAC_LANGID_MONGOLIAN_MONGOLIAN_SCRIPT' :  57,
    'TT_MAC_LANGID_MONGOLIAN_CYRILLIC_SCRIPT'  :  58,
    'TT_MAC_LANGID_PASHTO'                     :  59,
    'TT_MAC_LANGID_KURDISH'                    :  60,
    'TT_MAC_LANGID_KASHMIRI'                   :  61,
    'TT_MAC_LANGID_SINDHI'                     :  62,
    'TT_MAC_LANGID_TIBETAN'                    :  63,
    'TT_MAC_LANGID_NEPALI'                     :  64,
    'TT_MAC_LANGID_SANSKRIT'                   :  65,
    'TT_MAC_LANGID_MARATHI'                    :  66,
    'TT_MAC_LANGID_BENGALI'                    :  67,
    'TT_MAC_LANGID_ASSAMESE'                   :  68,
    'TT_MAC_LANGID_GUJARATI'                   :  69,
    'TT_MAC_LANGID_PUNJABI'                    :  70,
    'TT_MAC_LANGID_ORIYA'                      :  71,
    'TT_MAC_LANGID_MALAYALAM'                  :  72,
    'TT_MAC_LANGID_KANNADA'                    :  73,
    'TT_MAC_LANGID_TAMIL'                      :  74,
    'TT_MAC_LANGID_TELUGU'                     :  75,
    'TT_MAC_LANGID_SINHALESE'                  :  76,
    'TT_MAC_LANGID_BURMESE'                    :  77,
    'TT_MAC_LANGID_KHMER'                      :  78,
    'TT_MAC_LANGID_LAO'                        :  79,
    'TT_MAC_LANGID_VIETNAMESE'                 :  80,
    'TT_MAC_LANGID_INDONESIAN'                 :  81,
    'TT_MAC_LANGID_TAGALOG'                    :  82,
    'TT_MAC_LANGID_MALAY_ROMAN_SCRIPT'         :  83,
    'TT_MAC_LANGID_MALAY_ARABIC_SCRIPT'        :  84,
    'TT_MAC_LANGID_AMHARIC'                    :  85,
    'TT_MAC_LANGID_TIGRINYA'                   :  86,
    'TT_MAC_LANGID_GALLA'                      :  87,
    'TT_MAC_LANGID_SOMALI'                     :  88,
    'TT_MAC_LANGID_SWAHILI'                    :  89,
    'TT_MAC_LANGID_RUANDA'                     :  90,
    'TT_MAC_LANGID_RUNDI'                      :  91,
    'TT_MAC_LANGID_CHEWA'                      :  92,
    'TT_MAC_LANGID_MALAGASY'                   :  93,
    'TT_MAC_LANGID_ESPERANTO'                  :  94,
    'TT_MAC_LANGID_WELSH'                      : 128,
    'TT_MAC_LANGID_BASQUE'                     : 129,
    'TT_MAC_LANGID_CATALAN'                    : 130,
    'TT_MAC_LANGID_LATIN'                      : 131,
    'TT_MAC_LANGID_QUECHUA'                    : 132,
    'TT_MAC_LANGID_GUARANI'                    : 133,
    'TT_MAC_LANGID_AYMARA'                     : 134,
    'TT_MAC_LANGID_TATAR'                      : 135,
    'TT_MAC_LANGID_UIGHUR'                     : 136,
    'TT_MAC_LANGID_DZONGKHA'                   : 137,
    'TT_MAC_LANGID_JAVANESE'                   : 138,
    'TT_MAC_LANGID_SUNDANESE'                  : 139,
    'TT_MAC_LANGID_GALICIAN'                   : 140,
    'TT_MAC_LANGID_AFRIKAANS'                  : 141,
    'TT_MAC_LANGID_BRETON'                     : 142,
    'TT_MAC_LANGID_INUKTITUT'                  : 143,
    'TT_MAC_LANGID_SCOTTISH_GAELIC'            : 144,
    'TT_MAC_LANGID_MANX_GAELIC'                : 145,
    'TT_MAC_LANGID_IRISH_GAELIC'               : 146,
    'TT_MAC_LANGID_TONGAN'                     : 147,
    'TT_MAC_LANGID_GREEK_POLYTONIC'            : 148,
    'TT_MAC_LANGID_GREELANDIC'                 : 149,
    'TT_MAC_LANGID_AZERBAIJANI_ROMAN_SCRIPT'   : 150 }
globals().update(TT_MAC_LANGIDS)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
'''
Glyph colored rendering (with outline)
'''
from freetype import *

if __name__ == '__main__':
    import numpy as np
    import matplotlib.pyplot as plt

    face = Face('./Vera.ttf')
    face.set_char_size( 96*64 )
    RGBA = [('R',float), ('G',float), ('B',float), ('A',float)]

    # Outline
    flags = FT_LOAD_DEFAULT | FT_LOAD_NO_BITMAP
    face.load_char('S', flags )
    slot = face.glyph
    glyph = slot.get_glyph()
    stroker = Stroker( )
    stroker.set(64, FT_STROKER_LINECAP_ROUND, FT_STROKER_LINEJOIN_ROUND, 0 )
    glyph.stroke( stroker , True )
    blyph = glyph.to_bitmap(FT_RENDER_MODE_NORMAL, Vector(0,0), True )
    bitmap = blyph.bitmap
    width, rows, pitch = bitmap.width, bitmap.rows, bitmap.pitch
    top, left = blyph.top, blyph.left
    data = []
    for i in range(rows):
        data.extend(bitmap.buffer[i*pitch:i*pitch+width])
    Z = np.array(data).reshape(rows, width)/255.0
    O = np.zeros((rows,width), dtype=RGBA)
    O['A'] = Z
    O['R'] = 1
    O['G'] = 0
    O['B'] = 0

    # Plain
    flags = FT_LOAD_RENDER
    face.load_char('S', flags)
    F = np.zeros((rows,width), dtype=RGBA)
    Z = np.zeros((rows, width))
    bitmap = face.glyph.bitmap
    width, rows, pitch = bitmap.width, bitmap.rows, bitmap.pitch
    top, left = face.glyph.bitmap_top, face.glyph.bitmap_left
    dy = blyph.top - face.glyph.bitmap_top
    dx = face.glyph.bitmap_left - blyph.left
    data = []
    for i in range(rows):
        data.extend(bitmap.buffer[i*pitch:i*pitch+width])
    Z[dx:dx+rows,dy:dy+width] = np.array(data).reshape(rows, width)/255.
    F['R'] = 1
    F['G'] = 1
    F['B'] = 0
    F['A'] = Z

    # Combine outline and plain
    R1,G1,B1,A1 = O['R'],O['G'],O['B'],O['A']
    R2,G2,B2,A2 = F['R'],F['G'],F['B'],F['A']
    Z = np.zeros(O.shape, dtype=RGBA)
    Z['R'] = (A1 * R1 + A2 * (1 - A1) * R2)
    Z['G'] = (A1 * G1 + A2 * (1 - A1) * G2)
    Z['B'] = (A1 * B1 + A2 * (1 - A1) * B2)
    Z['A'] = (A1      + A2 * (1 - A1))


    # Draw
    plt.figure(figsize=(12,5))

    plt.subplot(1,3,1)
    plt.title('Plain')
    plt.xticks([]), plt.yticks([])
    I = F.view(dtype=float).reshape(O.shape[0],O.shape[1],4)
    plt.imshow(I, interpolation='nearest', origin='lower')

    plt.subplot(1,3,2)
    plt.title('Outline')
    plt.xticks([]), plt.yticks([])
    I = O.view(dtype=float).reshape(O.shape[0],O.shape[1],4)
    plt.imshow(I, interpolation='nearest', origin='lower')

    plt.subplot(1,3,3)
    plt.title('Outline + Plain')
    plt.xticks([]), plt.yticks([])
    I = Z.view(dtype=float).reshape(O.shape[0],O.shape[1],4)
    plt.imshow(I, interpolation='nearest', origin='lower')

    plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  pycairo/cairocffi-based glyph-vector-2 example - Copyright 2017 Hin-Tak Leung
#  Distributed under the terms of the new BSD license.
#
#  rewrite of the numply,matplotlib-based example from Nicolas P. Rougier
#  - The code is incomplete and over-simplified, as it ignores the 3rd order
#    bezier curve bit when intepolating between off-curve points.
#    This is only correct for truetype fonts (which only use 2nd order bezier curves).
#  - Also it seems to assume the first point is always on curve; this is
#    unusual but legal.
#
#  Can cope with well-behaved Postscript/CFF fonts too.
#
# -----------------------------------------------------------------------------
'''
Show how to access glyph outline description.
'''
from freetype import *
# using Matrix class from Cairo, instead of FreeType's
from cairo import Context, SVGSurface, Matrix, SurfacePattern, FILTER_BEST

from bitmap_to_surface import make_image_surface

if __name__ == '__main__':
    import numpy
    from PIL import Image

    # Replacement for Path enums:
    STOP, MOVETO, LINETO, CURVE3, CURVE4 = 0, 1, 2, 3, 4

    face = Face('./Vera.ttf')
    face.set_char_size( 32*64 )
    face.load_char('g')
    slot = face.glyph

    bitmap = face.glyph.bitmap
    width  = face.glyph.bitmap.width
    rows   = face.glyph.bitmap.rows
    pitch  = face.glyph.bitmap.pitch

    Z = make_image_surface(bitmap)

    outline = slot.outline
    points = numpy.array(outline.points, dtype=[('x',float), ('y',float)])
    x, y = points['x'], points['y']

    bbox = outline.get_cbox()

    MARGIN  = 10
    scale = 3

    def Floor64(x):
        return (x//64) * 64

    def Ceil64(x):
        return ((x+63)//64) * 64

    width_s = (width * 64)//scale + 2 * MARGIN
    height_s = (rows * 64)//scale + 2 * MARGIN

    surface = SVGSurface('glyph-vector-2-cairo.svg',
                         width_s,
                         height_s)
    ctx = Context(surface)
    ctx.set_source_rgb(1,1,1)
    ctx.paint()
    ctx.save()
    ctx.scale(1.0/scale,1.0/scale)
    ctx.translate(-Floor64(bbox.xMin) + MARGIN * scale,-Floor64(bbox.yMin) + MARGIN * scale)
    ctx.transform(Matrix(1,0,0,-1))
    ctx.translate(0, -(Ceil64(bbox.yMax) + Floor64(bbox.yMin))) # difference!

    start, end = 0, 0

    VERTS, CODES = [], []
    # Iterate over each contour
    for i in range(len(outline.contours)):
        end    = outline.contours[i]
        points = outline.points[start:end+1]
        points.append(points[0])
        tags   = outline.tags[start:end+1]
        tags.append(tags[0])

        segments = [ [points[0],], ]
        for j in range(1, len(points) ):
            segments[-1].append(points[j])
            if ( FT_Curve_Tag( tags[j] ) == FT_Curve_Tag_On ) and j < (len(points)-1):
                segments.append( [points[j],] )
        verts = [points[0], ]
        codes = [MOVETO,]
        tags.pop()
        for segment in segments:
            if len(segment) == 2:
                verts.extend(segment[1:])
                codes.extend([LINETO])
            elif len(segment) == 3:
                verts.extend(segment[1:])
                codes.extend([CURVE3, CURVE3])
            elif ( len(segment) == 4 ) \
                 and ( FT_Curve_Tag(tags[1]) == FT_Curve_Tag_Cubic ) \
                 and ( FT_Curve_Tag(tags[2]) == FT_Curve_Tag_Cubic ):
                verts.extend(segment[1:])
                codes.extend([CURVE4, CURVE4, CURVE4])
            else:
                # Interpolating
                verts.append(segment[1])
                codes.append(CURVE3)
                for i in range(1,len(segment)-2):
                    A,B = segment[i], segment[i+1]
                    C = ((A[0]+B[0])/2.0, (A[1]+B[1])/2.0)
                    verts.extend([ C, B ])
                    codes.extend([ CURVE3, CURVE3])
                verts.append(segment[-1])
                codes.append(CURVE3)
            [tags.pop() for x in range(len(segment) - 1)]
        VERTS.extend(verts)
        CODES.extend(codes)
        start = end+1


    # Draw glyph
    ctx.new_path()
    ctx.set_source_rgba(0.8,0.5,0.8, 1)
    i = 0
    while (i < len(CODES)):
        if (CODES[i] == MOVETO):
            ctx.move_to(VERTS[i][0],VERTS[i][1])
            i += 1
        elif (CODES[i] == LINETO):
            ctx.line_to(VERTS[i][0],VERTS[i][1])
            i += 1
        elif (CODES[i] == CURVE3):
            ctx.curve_to(VERTS[i][0],VERTS[i][1],
                         VERTS[i+1][0],VERTS[i+1][1], # undocumented
                         VERTS[i+1][0],VERTS[i+1][1])
            i += 2
        elif (CODES[i] == CURVE4):
            ctx.curve_to(VERTS[i][0],VERTS[i][1],
                         VERTS[i+1][0],VERTS[i+1][1],
                         VERTS[i+2][0],VERTS[i+2][1])
            i += 3
    ctx.fill_preserve()
    ctx.set_source_rgb(0,0,0)
    ctx.set_line_width(6)
    ctx.stroke()
    ctx.restore()

    scale2 = (height_s - 2.0 * MARGIN)/rows

    ctx.set_source_surface(Z, 0, 0)
    pattern = ctx.get_source()
    SurfacePattern.set_filter(pattern, FILTER_BEST)
    scalematrix = Matrix()
    scalematrix.scale(1.0/scale2, 1.0/scale2)
    scalematrix.translate(-( width_s/2.0  - width *scale2 /2.0 ), -MARGIN)
    pattern.set_matrix(scalematrix)
    ctx.set_source_rgba (0, 0, 0, 0.7)
    ctx.mask(pattern)
    ctx.fill()


    surface.flush()
    surface.write_to_png("glyph-vector-2-cairo.png")
    surface.finish()
    Image.open("glyph-vector-2-cairo.png").show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Corrected version of user-submitted code in https://github.com/rougier/freetype-py/issues/193

# Comments:
#     The original version missed "freetype.FT_LOAD_RENDER" and
#     "freetype.FT_LOAD_TARGET_MONO", and also has the unfortunate
#     setting of pixel size "16". Thus for some glyphs on 'simsun.ttc',
#     the code returns the embedded bitmap, while for other glyphs without
#     embedded bitmaps, it returns garbage for the bitmap buffer
#     (if you do not check "font.glyph.format" and "font.glyph.bitmap.pixel_mode" ;
#     font.glyph.format == FT_GLYPH_FORMAT_OUTLINE in that case). "bitmap.pitch" was also
#     not used in the original, and hard-coded length of scanline was used.

import freetype

font = freetype.Face(r"simsun.ttc")

# 绘制字符
font.set_pixel_sizes(19, 19)
font.load_char('1', freetype.FT_LOAD_RENDER | freetype.FT_LOAD_TARGET_MONO)
#font.load_char('字', freetype.FT_LOAD_RENDER | freetype.FT_LOAD_TARGET_MONO)
bitmap = font.glyph.bitmap
print(bitmap.rows, bitmap.width, bitmap.pitch)
assert(font.glyph.format == freetype.FT_GLYPH_FORMAT_BITMAP)
assert(font.glyph.bitmap.pixel_mode == freetype.FT_PIXEL_MODE_MONO)
print(bitmap.buffer)
print(len(bitmap.buffer))
left = []
right = []
for b, j in enumerate(bitmap.buffer):
    b_res = list(format(j, "08b"))
    for r in b_res:
        if r == "1":
            print("\033[1;30;46m   \033[0m", end="")
        else:
            print("\033[1;30;40m   \033[0m", end="")
    if (b + 1) % bitmap.pitch == 0 and b > 0:
        right.append(format(j, "02X"))
        print()
    else:
        left.append(format(j, "02X"))
print(' '.join(left))
print(' '.join(right))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  pycairo/cairocffi-based emoji-color example - Copyright 2017-2023 Hin-Tak Leung
#  Distributed under the terms of the new BSD license.
#
#  This script demonstrates overlapping emojis.
#
#  There are 4 kinds. sbix (Apple's) and CBDT/CBLC (Google's) only work
#  at specific sizes; OT-SVG (Adobe/Mozilla's) and COLR v0 (Microsoft's)
#  work at arbitrary sizes, but need the FT_LOAD_RENDER flags (this has
#  no effect on sbix and CBDT/CBLC, so can be added). It is added below
#  so this would work as is for COLR v0 (Microsoft's). In addition,
#  OT-SVG needs an external render. See ot-svg-example.py and comments
#  inside. Basically you need 3 more lines (rename "ot-svg-example.py" as
#  "otsvg.py", then add 3 lines "from otsvg import hooks",
#  get library handle, set the "ot-svg" "svg-hooks" property, to load it)
#
#  Note: On Mac OS X before Sierra (10.12), change ttc->ttf;
#        try Google's NotoColorEmoji.ttf at size 109 on Linux.
#
#  Updated 2023, now somewhat specific to pycairo and CPython 3.3+,
#  and also small-endian (from FORMAT_ARGB32 vs FT_PIXEL_MODE_BGRA),
#  but no longer needs numpy.
#
#  Older 2017 version (depends on numpy):
#  Limitation: Suface.get_data() is not in the "python 3, pycairo < 1.11" combo.

import freetype
from PIL import Image

from cairo import ImageSurface, FORMAT_ARGB32, Context

face = freetype.Face("/System/Library/Fonts/Apple Color Emoji.ttc")
# Not all char sizes are valid for emoji fonts;
# Google's NotoColorEmoji only accept size 109 to get 136x128 bitmaps
face.set_char_size( 160*64 )
face.load_char('😀', freetype.FT_LOAD_COLOR | freetype.FT_LOAD_RENDER)

bitmap = face.glyph.bitmap
width = face.glyph.bitmap.width
rows = face.glyph.bitmap.rows

if ( face.glyph.bitmap.pitch != width * 4 ):
    raise RuntimeError('pitch != width * 4 for color bitmap: Please report this.')

# See https://stackoverflow.com/questions/59574816/obtaining-a-memoryview-object-from-a-ctypes-c-void-p-object
from ctypes import pythonapi, c_char_p, c_ssize_t, c_int, py_object, cast
pythonapi.PyMemoryView_FromMemory.argtypes = (c_char_p, c_ssize_t, c_int)
pythonapi.PyMemoryView_FromMemory.restype = py_object
I = ImageSurface.create_for_data( pythonapi.PyMemoryView_FromMemory(cast(bitmap._FT_Bitmap.buffer, c_char_p),
                                                                    bitmap.rows * bitmap.pitch,
                                                                    0x200), # Read-Write
                                        FORMAT_ARGB32,
                                        width, rows,
                                        bitmap.pitch )
surface = ImageSurface(FORMAT_ARGB32, 2*width, rows)
ctx = Context(surface)

ctx.set_source_surface(I, 0, 0)
ctx.paint()

ctx.set_source_surface(I, width/2, 0)
ctx.paint()

ctx.set_source_surface(I, width , 0)
ctx.paint()

surface.write_to_png("emoji-color-cairo.png")
Image.open("emoji-color-cairo.png").show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
from freetype import *
import numpy as np
from PIL import Image


def render(filename = "Vera.ttf", hinting = (False,False), gamma = 1.5, lcd=False):
    text = "A Quick Brown Fox Jumps Over The Lazy Dog 0123456789"

    W,H,D = 680, 280, 1
    Z = np.zeros( (H,W), dtype=np.ubyte )
    face = Face(filename)
    pen = Vector(5*64, (H-10)*64)

    flags = FT_LOAD_RENDER
    if hinting[1]: flags |= FT_LOAD_FORCE_AUTOHINT
    else:          flags |= FT_LOAD_NO_HINTING
    if hinting[0]: hres, hscale = 72,    1.0
    else:          hres, hscale = 72*10, 0.1
    if lcd:
        flags |= FT_LOAD_TARGET_LCD
        Z = np.zeros( (H,W,3), dtype=np.ubyte )
        try:
            set_lcd_filter( FT_LCD_FILTER_DEFAULT )
        except:
            pass


    for size in range(9,23):
        face.set_char_size( size * 64, 0, hres, 72 )
        matrix = Matrix( int((hscale) * 0x10000), int((0.0) * 0x10000),
                         int((0.0)    * 0x10000), int((1.0) * 0x10000) )
        previous = 0
        pen.x = 5*64
        for current in text:
            face.set_transform( matrix, pen )
            face.load_char( current, flags)
            kerning = face.get_kerning( previous, current, FT_KERNING_UNSCALED )
            pen.x += kerning.x
            glyph = face.glyph
            bitmap = glyph.bitmap
            x, y = glyph.bitmap_left, glyph.bitmap_top
            w, h, p = bitmap.width, bitmap.rows, bitmap.pitch
            buff = np.array(bitmap.buffer, dtype=np.ubyte).reshape((h,p))
            if lcd:
                Z[H-y:H-y+h,x:x+w/3].flat |= buff[:,:w].flatten()
            else:
                Z[H-y:H-y+h,x:x+w].flat |= buff[:,:w].flatten()
            pen.x += glyph.advance.x
            previous = current
        pen.y -= (size+4)*64

    # Gamma correction
    Z = (Z/255.0)**(gamma)
    Z = ((1-Z)*255).astype(np.ubyte)
    if lcd:
        I = Image.fromarray(Z, mode='RGB')
    else:
        I = Image.fromarray(Z, mode='L')

    name = filename.split('.')[0]
    filename = '%s-gamma(%.1f)-hinting(%d,%d)-lcd(%d).png' % (name,gamma,hinting[0],hinting[1],lcd)
    I.save(filename)



if __name__ == '__main__':
    render('Vera.ttf', (0,1), 1.25, True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  OT-SVG example with Skia
#
#  Copyright 2023 Hin-Tak Leung
#  Distributed under the terms of the new BSD license.

# This is an older example which works on skia m87 with known limitations.
# See "ot-svg-example-skia.py" for a full example which requires m117.

# Limitation:
#     Skia-python bundles with Skia m87 (at time of writing this).
#
#     Skia m88 is first version where SkSVG* is considered no longer experimental.
#
#     Skia m103 is the first Skia build which contains 9cbadcd9280dc139af2f4d41d25a6c9a750e0302.
#     That introduces "SkSVGDOM::renderNode()" among other stuff,
#     necessary for rendering "some" OT-SVG fonts. Guess what, that commit
#     is titled "Add optional OT-SVG support to FreeType"!
#
#     So the example below only works correctly for "some" glyphs in
#     "some other" OT-SVG fonts, and also with very limited functionality
#     beyond what is used below.
#
#     The missing functionality (and support for beyond Skia m103) is filed
#     as skia-python issue #192.

from freetype import *

import skia

if __name__ == '__main__':
    import sys
    execname = sys.argv[0]

    if len(sys.argv) < 2:
        print("Example usage: %s TrajanColor-Concept.otf" % execname)
        exit(1)

    face = Face(sys.argv[1])

    face.set_char_size( 160*64 )
    face.load_char('A', FT_LOAD_COLOR )
    slot = face.glyph._FT_GlyphSlot

    if (face.glyph.format == FT_GLYPH_FORMAT_SVG):
        document = ctypes.cast(slot.contents.other, FT_SVG_Document)
        doc = ctypes.string_at(document.contents.svg_document, # not terminated
                               size=document.contents.svg_document_length)
        d = skia.Data(doc)
        # skia-python #194 - the raw c++ api actually has a (ptr, len, bool) method!
        # Data() shouldn't be needed.
        m = skia.MemoryStream(d)
        h = skia.SVGDOM.MakeFromStream(m)

        WIDTH, HEIGHT = 160, 160
        
        size = skia.Size()
        size.fHeight = WIDTH
        size.fWidth = HEIGHT
        
        h.setContainerSize(size)

        surface = skia.Surface(WIDTH * 2, HEIGHT)
        
        with surface as canvas:
            canvas.translate(0,HEIGHT)
            h.render(canvas)
            canvas.translate(WIDTH/4,0)
            h.render(canvas)
            canvas.translate(WIDTH/4,0)
            h.render(canvas)
            canvas.translate(WIDTH/4,0)
            h.render(canvas)
            canvas.translate(WIDTH/4,0)
            h.render(canvas)
            canvas.translate(WIDTH/4,0)
            h.render(canvas)
            
        surface.flushAndSubmit()
        image = surface.makeImageSnapshot()
        image.save("ot-svg-draw-skia.png", skia.kPNG)

        from PIL import Image
        Image.open("ot-svg-draw-skia.png").show()

    else:
        print("Not SVG glyph.format")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
'''
Subpixel rendering AND positioning using OpenGL and shaders.

'''
import numpy as np
import OpenGL.GL as gl
import OpenGL.GLUT as glut
from texture_font import TextureFont, TextureAtlas
from shader import Shader


vert='''
uniform sampler2D texture;
uniform vec2 pixel;
attribute float modulo;
varying float m;
void main() {
    gl_FrontColor = gl_Color;
    gl_TexCoord[0].xy = gl_MultiTexCoord0.xy;
    gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;
    m = modulo;
}
'''

frag='''
uniform sampler2D texture;
uniform vec2 pixel;
varying float m;
void main() {
    float gamma = 1.0;

    vec2 uv      = gl_TexCoord[0].xy;
    vec4 current = texture2D(texture, uv);
    vec4 previous= texture2D(texture, uv+vec2(-1,0)*pixel);

    current  = pow(current,  vec4(1.0/gamma));
    previous = pow(previous, vec4(1.0/gamma));

    float r = current.r;
    float g = current.g;
    float b = current.b;
    float a = current.a;
    if( m <= 0.333 )
    {
        float z = m/0.333;
        r = mix(current.r, previous.b, z);
        g = mix(current.g, current.r,  z);
        b = mix(current.b, current.g,  z);
    }
    else if( m <= 0.666 )
    {
        float z = (m-0.33)/0.333;
        r = mix(previous.b, previous.g, z);
        g = mix(current.r,  previous.b, z);
        b = mix(current.g,  current.r,  z);
    }
   else if( m < 1.0 )
    {
        float z = (m-0.66)/0.334;
        r = mix(previous.g, previous.r, z);
        g = mix(previous.b, previous.g, z);
        b = mix(current.r,  previous.b, z);
    }

   float t = max(max(r,g),b);
   vec4 color = vec4(0.,0.,0., (r+g+b)/2.);
   color = t*color + (1.-t)*vec4(r,g,b, min(min(r,g),b));
   gl_FragColor = vec4( color.rgb, color.a);
}
'''





class Label:
    def __init__(self, text, font, color=(1.0, 1.0, 1.0, 0.0),  x=0, y=0,
                 width=None, height=None, anchor_x='left', anchor_y='baseline'):
        self.text = text
        self.vertices = np.zeros((len(text)*4,3), dtype=np.float32)
        self.indices  = np.zeros((len(text)*6, ), dtype=np.uint)
        self.colors   = np.zeros((len(text)*4,4), dtype=np.float32)
        self.texcoords= np.zeros((len(text)*4,2), dtype=np.float32)
        self.attrib   = np.zeros((len(text)*4,1), dtype=np.float32)
        pen = [x,y]
        prev = None

        for i,charcode in enumerate(text):
            glyph = font[charcode]
            kerning = glyph.get_kerning(prev)
            x0 = pen[0] + glyph.offset[0] + kerning
            dx = x0-int(x0)
            x0 = int(x0)
            y0 = pen[1] + glyph.offset[1]
            x1 = x0 + glyph.size[0]
            y1 = y0 - glyph.size[1]
            u0 = glyph.texcoords[0]
            v0 = glyph.texcoords[1]
            u1 = glyph.texcoords[2]
            v1 = glyph.texcoords[3]

            index     = i*4
            indices   = [index, index+1, index+2, index, index+2, index+3]
            vertices  = [[x0,y0,1],[x0,y1,1],[x1,y1,1], [x1,y0,1]]
            texcoords = [[u0,v0],[u0,v1],[u1,v1], [u1,v0]]
            colors    = [color,]*4

            self.vertices[i*4:i*4+4] = vertices
            self.indices[i*6:i*6+6] = indices
            self.texcoords[i*4:i*4+4] = texcoords
            self.colors[i*4:i*4+4] = colors
            self.attrib[i*4:i*4+4] = dx
            pen[0] = pen[0]+glyph.advance[0]/64.0 + kerning
            pen[1] = pen[1]+glyph.advance[1]/64.0
            prev = charcode

        width = pen[0]-glyph.advance[0]/64.0+glyph.size[0]

        if anchor_y == 'top':
            dy = -round(font.ascender)
        elif anchor_y == 'center':
            dy = +round(-font.height/2-font.descender)
        elif anchor_y == 'bottom':
            dy = -round(font.descender)
        else:
            dy = 0

        if anchor_x == 'right':
            dx = -width/1.0
        elif anchor_x == 'center':
            dx = -width/2.0
        else:
            dx = 0
        self.vertices += (round(dx), round(dy), 0)


    def draw(self):
        gl.glEnable( gl.GL_TEXTURE_2D )
        gl.glDisable( gl.GL_DEPTH_TEST )

        gl.glEnableClientState(gl.GL_VERTEX_ARRAY)
        gl.glEnableClientState(gl.GL_COLOR_ARRAY)
        gl.glEnableClientState(gl.GL_TEXTURE_COORD_ARRAY)
        gl.glEnableClientState(gl.GL_VERTEX_ARRAY)

        gl.glVertexPointer(3, gl.GL_FLOAT, 0, self.vertices)
        gl.glColorPointer(4, gl.GL_FLOAT, 0, self.colors)
        gl.glTexCoordPointer(2, gl.GL_FLOAT, 0, self.texcoords)

        r,g,b = 0,0,0
        gl.glColor( 1, 1, 1, 1 )
        gl.glEnable( gl.GL_BLEND )
        #gl.glBlendFunc( gl.GL_CONSTANT_COLOR_EXT,  gl.GL_ONE_MINUS_SRC_COLOR )
        #gl.glBlendColor(r,g,b,1)
        gl.glBlendFunc( gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA )
        gl.glBlendColor( 1, 1, 1, 1 )

        gl.glEnableVertexAttribArray( 1 );
        gl.glVertexAttribPointer( 1, 1, gl.GL_FLOAT, gl.GL_FALSE, 0, self.attrib)
        shader.bind()
        shader.uniformi('texture', 0)
        shader.uniformf('pixel', 1.0/512, 1.0/512)
        gl.glDrawElements(gl.GL_TRIANGLES, len(self.indices),
                          gl.GL_UNSIGNED_INT, self.indices)
        shader.unbind()
        gl.glDisableVertexAttribArray( 1 );
        gl.glDisableClientState(gl.GL_VERTEX_ARRAY)
        gl.glDisableClientState(gl.GL_COLOR_ARRAY)
        gl.glDisableClientState(gl.GL_TEXTURE_COORD_ARRAY)
        gl.glDisable( gl.GL_TEXTURE_2D )
        gl.glDisable( gl.GL_BLEND )




if __name__ == '__main__':
    import sys

    atlas = TextureAtlas(512,512,3)

    def on_display( ):
        #gl.glClearColor(0,0,0,1)
        gl.glClearColor(1,1,1,1)
        gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
        gl.glBindTexture( gl.GL_TEXTURE_2D, atlas.texid )
        for label in labels:
            label.draw()

        gl.glColor(0,0,0,1)
        gl.glBegin(gl.GL_LINES)
        gl.glVertex2i(15,0)
        gl.glVertex2i(15, 330)
        gl.glVertex2i(225, 0)
        gl.glVertex2i(225, 330)
        gl.glEnd()
        glut.glutSwapBuffers( )

    def on_reshape( width, height ):
        gl.glViewport( 0, 0, width, height )
        gl.glMatrixMode( gl.GL_PROJECTION )
        gl.glLoadIdentity( )
        gl.glOrtho( 0, width, 0, height, -1, 1 )
        gl.glMatrixMode( gl.GL_MODELVIEW )
        gl.glLoadIdentity( )

    def on_keyboard( key, x, y ):
        if key == '\033':
            sys.exit( )

    glut.glutInit( sys.argv )
    glut.glutInitDisplayMode( glut.GLUT_DOUBLE | glut.GLUT_RGBA | glut.GLUT_DEPTH )
    glut.glutCreateWindow( "Freetype OpenGL" )
    glut.glutReshapeWindow( 240, 330 )
    glut.glutDisplayFunc( on_display )
    glut.glutReshapeFunc( on_reshape )
    glut.glutKeyboardFunc( on_keyboard )

    font = TextureFont(atlas, './Vera.ttf', 9)
    text = "|... A Quick Brown Fox Jumps Over The Lazy Dog"
    labels = []
    x,y = 20,310
    for i in range(30):
        labels.append(Label(text=text, font=font, x=x, y=y))
        x += 0.1000000000001
        y -= 10
    atlas.upload()
    shader = Shader(vert,frag)
    glut.glutMainLoop( )



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
"""
An enumeration type that lists the render modes supported by FreeType 2. Each
mode corresponds to a specific type of scanline conversion performed on the
outline.

For bitmap fonts and embedded bitmaps the 'bitmap->pixel_mode' field in the
FT_GlyphSlotRec structure gives the format of the returned bitmap.

All modes except FT_RENDER_MODE_MONO use 256 levels of opacity.


FT_RENDER_MODE_NORMAL

  This is the default render mode; it corresponds to 8-bit anti-aliased
  bitmaps.


FT_RENDER_MODE_LIGHT

  This is equivalent to FT_RENDER_MODE_NORMAL. It is only defined as a separate
  value because render modes are also used indirectly to define hinting
  algorithm selectors. See FT_LOAD_TARGET_XXX for details.


FT_RENDER_MODE_MONO

  This mode corresponds to 1-bit bitmaps (with 2 levels of opacity).


FT_RENDER_MODE_LCD

  This mode corresponds to horizontal RGB and BGR sub-pixel displays like LCD
  screens. It produces 8-bit bitmaps that are 3 times the width of the original
  glyph outline in pixels, and which use the FT_PIXEL_MODE_LCD mode.


FT_RENDER_MODE_LCD_V

  This mode corresponds to vertical RGB and BGR sub-pixel displays (like PDA
  screens, rotated LCD displays, etc.). It produces 8-bit bitmaps that are 3
  times the height of the original glyph outline in pixels and use the
  FT_PIXEL_MODE_LCD_V mode.
  
FT_RENDER_MODE_SDF

  This mode corresponds to 8-bit, single-channel signed distance field (SDF)
  bitmaps. Each pixel in the SDF grid is the value from the pixel's position to
  the nearest glyph's outline. The distances are calculated from the center of
  the pixel and are positive if they are filled by the outline (i.e., inside
  the outline) and negative otherwise. Check the note below on how to convert
  the output values to usable data.
"""
FT_RENDER_MODES = { 'FT_RENDER_MODE_NORMAL' : 0,
                    'FT_RENDER_MODE_LIGHT'  : 1,
                    'FT_RENDER_MODE_MONO'   : 2,
                    'FT_RENDER_MODE_LCD'    : 3,
                    'FT_RENDER_MODE_LCD_V'  : 4,
                    'FT_RENDER_MODE_SDF'    : 5 }
globals().update(FT_RENDER_MODES)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  pycairo/cairocffi-based glyph-color example - Copyright 2017 Hin-Tak Leung
#  Distributed under the terms of the new BSD license.
#
#  rewrite of the numply,matplotlib-based example from Nicolas P. Rougier
#  - The two side-glyphs are positioned only roughly equivalently.
#
# -----------------------------------------------------------------------------
'''
Glyph colored rendering (with outline)
'''
from freetype import *

# using Matrix class from Cairo, instead of FreeType's
from cairo import Context, ImageSurface, FORMAT_ARGB32, SurfacePattern, FILTER_BEST, Matrix
from bitmap_to_surface import make_image_surface

if __name__ == '__main__':
    from PIL import Image

    face = Face('./Vera.ttf')
    face.set_char_size( 96*64 )

    # Outline
    flags = FT_LOAD_DEFAULT | FT_LOAD_NO_BITMAP
    face.load_char('S', flags )
    slot = face.glyph
    glyph = slot.get_glyph()
    stroker = Stroker( )
    stroker.set(64, FT_STROKER_LINECAP_ROUND, FT_STROKER_LINEJOIN_ROUND, 0 )
    glyph.stroke( stroker , True )
    del stroker
    blyph = glyph.to_bitmap(FT_RENDER_MODE_NORMAL, Vector(0,0), True )
    bitmap = blyph.bitmap
    widthZ, rowsZ, pitch = bitmap.width, bitmap.rows, bitmap.pitch
    Z = make_image_surface(bitmap)

    # Plain
    flags = FT_LOAD_RENDER
    face.load_char('S', flags)
    bitmap = face.glyph.bitmap
    widthF, rowsF, pitch = bitmap.width, bitmap.rows, bitmap.pitch
    F = make_image_surface(bitmap)

    # Draw
    surface = ImageSurface(FORMAT_ARGB32, 1200, 500)
    ctx = Context(surface)

    # fill background as gray
    ctx.rectangle(0,0,1200,500)
    ctx.set_source_rgb (0.5 , 0.5, 0.5)
    ctx.fill()

    # use the stroked font's size as scale, as it is likely slightly larger
    scale = 400.0 / rowsZ

    # draw bitmap first
    ctx.set_source_surface(F, 0, 0)
    patternF = ctx.get_source()
    SurfacePattern.set_filter(patternF, FILTER_BEST)

    scalematrix = Matrix()
    scalematrix.scale(1.0/scale,1.0/scale)
    scalematrix.translate(-(600.0 - widthF *scale /2.0 ), -50)
    patternF.set_matrix(scalematrix)
    ctx.set_source_rgb (1 , 1, 0)
    ctx.mask(patternF)
    ctx.fill()

    scalematrix.translate(+400,0)
    patternF.set_matrix(scalematrix)
    ctx.mask(patternF)
    ctx.fill()

    # stroke on top
    ctx.set_source_surface(Z, 0, 0)
    patternZ = ctx.get_source()
    SurfacePattern.set_filter(patternZ, FILTER_BEST)

    scalematrix = Matrix()
    scalematrix.scale(1.0/scale,1.0/scale)
    scalematrix.translate(-(600.0 - widthZ *scale /2.0 ), -50)
    patternZ.set_matrix(scalematrix)
    ctx.set_source_rgb (1 , 0, 0)
    ctx.mask(patternZ)
    ctx.fill()

    scalematrix.translate(-400,0)
    patternZ.set_matrix(scalematrix)
    ctx.mask(patternZ)
    ctx.fill()

    surface.flush()
    surface.write_to_png("glyph-color-cairo.png")
    Image.open("glyph-color-cairo.png").show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
from freetype import *

if __name__ == '__main__':
    import sys

    if len(sys.argv) < 2:
        print("Usage: %s font_filename" % sys.argv[0])
        sys.exit()

    face = Face(sys.argv[1])

    print ('Family name:         {}'.format(face.family_name.decode("ascii", errors='ignore')))
    print ('Style name:          {}'.format(face.style_name.decode("ascii", errors='ignore')))
    print ('Charmaps:            {}'.format([charmap.encoding_name for charmap in face.charmaps]))
    print ('')
    print ('Face number:         {}'.format(face.num_faces))
    print ('Glyph number:        {}'.format(face.num_glyphs))
    # FT_Bitmap_Size.size is in F26.6. Need to divide by 64:
    print ('Available sizes:     {}'.format([bitmap_size.size/64 for bitmap_size in face.available_sizes]))
    print ('')
    print ('units per em:        {}'.format(face.units_per_EM))
    print ('ascender:            {}'.format(face.ascender))
    print ('descender:           {}'.format(face.descender))
    print ('height:              {}'.format(face.height))
    print ('')
    print ('max_advance_width:   {}'.format(face.max_advance_width))
    print ('max_advance_height:  {}'.format(face.max_advance_height))
    print ('')
    print ('underline_position:  {}'.format(face.underline_position))
    print ('underline_thickness: {}'.format(face.underline_thickness))
    print ('')
    print ('Has horizontal:      {}'.format(face.has_horizontal))
    print ('Has vertical:        {}'.format(face.has_vertical))
    print ('Has kerning:         {}'.format(face.has_kerning))
    print ('Is fixed width:      {}'.format(face.is_fixed_width))
    print ('Is scalable:         {}'.format(face.is_scalable))
    print ('')



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
'''
Internal exception with freetype error message
'''
class FT_Exception(Exception):
    def __init__(self, errcode, message=''):
        self.message = message
        self.errcode = errcode

    def __str__(self):
        return '%s: %s (%s)'%(self.__class__.__name__, self.message,
            self._errors.get(self.errcode, 'unknown error'))

    _errors = {
        0x00: "no error" ,
        0x01: "cannot open resource" ,
        0x02: "unknown file format" ,
        0x03: "broken file" ,
        0x04: "invalid FreeType version" ,
        0x05: "module version is too low" ,
        0x06: "invalid argument" ,
        0x07: "unimplemented feature" ,
        0x08: "broken table" ,
        0x09: "broken offset within table" ,
        0x0A: "array allocation size too large" ,
        0x0B: "missing module" ,
        0x0C: "missing property" ,
        0x10: "invalid glyph index" ,
        0x11: "invalid character code" ,
        0x12: "unsupported glyph image format" ,
        0x13: "cannot render this glyph format" ,
        0x14: "invalid outline" ,
        0x15: "invalid composite glyph" ,
        0x16: "too many hints" ,
        0x17: "invalid pixel size" ,
        0x18: "invalid SVG document" ,
        0x20: "invalid object handle" ,
        0x21: "invalid library handle" ,
        0x22: "invalid module handle" ,
        0x23: "invalid face handle" ,
        0x24: "invalid size handle" ,
        0x25: "invalid glyph slot handle" ,
        0x26: "invalid charmap handle" ,
        0x27: "invalid cache manager handle" ,
        0x28: "invalid stream handle" ,
        0x30: "too many modules" ,
        0x31: "too many extensions" ,
        0x40: "out of memory" ,
        0x41: "unlisted object" ,
        0x51: "cannot open stream" ,
        0x52: "invalid stream seek" ,
        0x53: "invalid stream skip" ,
        0x54: "invalid stream read" ,
        0x55: "invalid stream operation" ,
        0x56: "invalid frame operation" ,
        0x57: "nested frame access" ,
        0x58: "invalid frame read" ,
        0x60: "raster uninitialized" ,
        0x61: "raster corrupted" ,
        0x62: "raster overflow" ,
        0x63: "negative height while rastering" ,
        0x70: "too many registered caches" ,
        0x80: "invalid opcode" ,
        0x81: "too few arguments" ,
        0x82: "stack overflow" ,
        0x83: "code overflow" ,
        0x84: "bad argument" ,
        0x85: "division by zero" ,
        0x86: "invalid reference" ,
        0x87: "found debug opcode" ,
        0x88: "found ENDF opcode in execution stream" ,
        0x89: "nested DEFS" ,
        0x8A: "invalid code range" ,
        0x8B: "execution context too long" ,
        0x8C: "too many function definitions" ,
        0x8D: "too many instruction definitions" ,
        0x8E: "SFNT font table missing" ,
        0x8F: "horizontal header (hhea, table missing" ,
        0x90: "locations (loca, table missing" ,
        0x91: "name table missing" ,
        0x92: "character map (cmap, table missing" ,
        0x93: "horizontal metrics (hmtx, table missing" ,
        0x94: "PostScript (post, table missing" ,
        0x95: "invalid horizontal metrics" ,
        0x96: "invalid character map (cmap, format" ,
        0x97: "invalid ppem value" ,
        0x98: "invalid vertical metrics" ,
        0x99: "could not find context" ,
        0x9A: "invalid PostScript (post, table format" ,
        0x9B: "invalid PostScript (post, table" ,
        0x9C: "found FDEF or IDEF opcode in glyf bytecode" ,
        0x9D: "missing bitmap in strike" ,
        0x9E: "SVG hooks have not been set",
        0xA0: "opcode syntax error" ,
        0xA1: "argument stack underflow" ,
        0xA2: "ignore" ,
        0xA3: "no Unicode glyph name found" ,
        0xA4: "glyph too big for hinting" ,
        0xB0: "`STARTFONT' field missing" ,
        0xB1: "`FONT' field missing" ,
        0xB2: "`SIZE' field missing" ,
        0xB3: "`FONTBOUNDINGBOX' field missing" ,
        0xB4: "`CHARS' field missing" ,
        0xB5: "`STARTCHAR' field missing" ,
        0xB6: "`ENCODING' field missing" ,
        0xB7: "`BBX' field missing" ,
        0xB8: "`BBX' too big" ,
        0xB9: "Font header corrupted or missing fields" ,
        0xBA: "Font glyphs corrupted or missing fields" ,
    }

FT_Err_Ok=0x00



# FT_Bitmap to CAIRO_SURFACE_TYPE_IMAGE module
# ============================================
#
# Copyright 2017 Hin-Tak Leung
#
# FreeType is under FTL (BSD license with an advertising clause) or GPLv2+.
# Cairo is under LGPLv2 or MPLv1.1.
#
# This is a heavily modified copy of a few routines from Lawrence D'Oliveiro[1],
# adjusted for freetype-py, and bugfix/workaround for mono-rendering [2].
#
# The bugfix/workaround requires libtiff on small-endian platforms.
#
# TODO: Look into using FreeType's FT_Bitmap_Convert() instead. However, libtiff
#      is common enough, and probably not important.
#
#[1] https://github.com/ldo/python_freetype
#    https://github.com/ldo/python_freetype_examples
#
#[2] https://github.com/ldo/python_freetype/issues/1
#    https://github.com/ldo/python_freetype_examples/issues/1
#
'''
FT_Bitmap to CAIRO_SURFACE_TYPE_IMAGE module
============================================

Converting from Freetype's FT_Bitmap to Cairo's CAIRO_SURFACE_TYPE_IMAGE

Usage:
     from bitmap_to_surface import make_image_surface

Works with cairocffi too. (Replace "from cairo ..." with "from cairocffi ...")

Limitation: Surface.create_for_data is not in the "python 3, pycairo < 1.11" combo.
'''
from freetype import FT_PIXEL_MODE_MONO, FT_PIXEL_MODE_GRAY, FT_PIXEL_MODE_BGRA, FT_Pointer, FT_Bitmap

from cairo import ImageSurface, FORMAT_A1, FORMAT_A8, FORMAT_ARGB32
#from cairocffi import ImageSurface, FORMAT_A1, FORMAT_A8, FORMAT_ARGB32

from array import array
from ctypes import cast, memmove, CDLL, c_void_p, c_int
from sys import byteorder

def make_image_surface(bitmap, copy = True) :
    if ( type(bitmap) == FT_Bitmap ):
        # bare FT_Bitmap
        content = bitmap
    else:
        # wrapped Bitmap instance
        content = bitmap._FT_Bitmap
    "creates a Cairo ImageSurface containing (a copy of) the Bitmap pixels."
    if content.pixel_mode == FT_PIXEL_MODE_MONO :
        cairo_format = FORMAT_A1
    elif content.pixel_mode == FT_PIXEL_MODE_GRAY :
        cairo_format = FORMAT_A8
    elif content.pixel_mode == FT_PIXEL_MODE_BGRA : # small-endian
        cairo_format = FORMAT_ARGB32 # platform native
    else :
        raise NotImplementedError("unsupported bitmap format %d" % content.pixel_mode)
    src_pitch = content.pitch
    dst_pitch = ImageSurface.format_stride_for_width(cairo_format, content.width)
    if not copy and dst_pitch == src_pitch and content.buffer != None :
        pixels = content.buffer
    else :
        pixels = to_array(content, content.pixel_mode, dst_pitch)
    result =  ImageSurface.create_for_data(
        pixels, cairo_format,
        content.width, content.rows,
        dst_pitch)
    return result

def to_array(content, pixel_mode, dst_pitch = None) :
    "returns a Python array object containing a copy of the Bitmap pixels."
    if dst_pitch == None :
        dst_pitch = content.pitch
    buffer_size = content.rows * dst_pitch
    buffer = array("B", b"0" * buffer_size)
    dstaddr = buffer.buffer_info()[0]
    srcaddr = cast(content.buffer, FT_Pointer).value
    src_pitch = content.pitch
    if dst_pitch == src_pitch :
        memmove(dstaddr, srcaddr, buffer_size)
    else :
        # have to copy a row at a time
        if src_pitch < 0 or dst_pitch < 0 :
            raise NotImplementedError("can't cope with negative bitmap pitch")
        assert dst_pitch > src_pitch
        for i in range(content.rows) :
            memmove(dstaddr, srcaddr, src_pitch)
            dstaddr += dst_pitch
            srcaddr += src_pitch
    # pillow/PIL itself requires libtiff so it is assumed to be around.
    # swap the bit-order from freetype's (MSB) to cairo's (host order) if needed
    if ( ( byteorder == 'little' ) and (pixel_mode == FT_PIXEL_MODE_MONO ) ):
        libtiff = CDLL("libtiff.so.5")
        libtiff.TIFFReverseBits.restype = None
        libtiff.TIFFReverseBits.argtypes = (c_void_p, c_int)
        libtiff.TIFFReverseBits(buffer.buffer_info()[0], buffer_size)
    return buffer



"""Custom PEP 517 build backend to provide different dependencies for wheel builds.

We need some extra dependencies when we build a wheel that bundles the FreeType
shared library (triggered when FREETYPEPY_BUNDLE_FT environment variable is set),
as opposed to a pure Python wheel (py3-none-any.whl).

For more info see:

https://setuptools.pypa.io/en/latest/build_meta.html#dynamic-build-dependencies-and-other-build-meta-tweaks

https://github.com/rougier/freetype-py/issues/183
"""

import os
from setuptools import build_meta as _orig
from setuptools.build_meta import *


def get_requires_for_build_wheel(config_settings=None):
    build_wheel_deps = _orig.get_requires_for_build_wheel(config_settings)

    if os.environ.get("FREETYPEPY_BUNDLE_FT"):
        build_wheel_deps += ["certifi", "cmake"]

    return build_wheel_deps


def get_requires_for_build_editable(config_settings=None):
    # ensure pip install -e . uses same build deps as regular pip install/wheel
    return get_requires_for_build_wheel(config_settings)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
"""
A list of values that are used to select a specific hinting algorithm to use
by the hinter. You should OR one of these values to your 'load_flags' when
calling FT_Load_Glyph.

Note that font's native hinters may ignore the hinting algorithm you have
specified (e.g., the TrueType bytecode interpreter). You can set
FT_LOAD_FORCE_AUTOHINT to ensure that the auto-hinter is used.

Also note that FT_LOAD_TARGET_LIGHT is an exception, in that it always
implies FT_LOAD_FORCE_AUTOHINT.


FT_LOAD_TARGET_NORMAL

  This corresponds to the default hinting algorithm, optimized for standard
  gray-level rendering. For monochrome output, use FT_LOAD_TARGET_MONO instead.


FT_LOAD_TARGET_LIGHT

  A lighter hinting algorithm for non-monochrome modes. Many generated glyphs
  are more fuzzy but better resemble its original shape. A bit like rendering
  on Mac OS X.

  As a special exception, this target implies FT_LOAD_FORCE_AUTOHINT.


FT_LOAD_TARGET_MONO

  Strong hinting algorithm that should only be used for monochrome output. The
  result is probably unpleasant if the glyph is rendered in non-monochrome
  modes.


FT_LOAD_TARGET_LCD

  A variant of FT_LOAD_TARGET_NORMAL optimized for horizontally decimated LCD
  displays.


FT_LOAD_TARGET_LCD_V

  A variant of FT_LOAD_TARGET_NORMAL optimized for vertically decimated LCD
  displays.
"""

from freetype.ft_enums.ft_render_modes import *


def _FT_LOAD_TARGET_(x):
    return (x & 15) << 16
FT_LOAD_TARGETS = {
    'FT_LOAD_TARGET_NORMAL' : _FT_LOAD_TARGET_(FT_RENDER_MODE_NORMAL),
    'FT_LOAD_TARGET_LIGHT'  : _FT_LOAD_TARGET_(FT_RENDER_MODE_LIGHT),
    'FT_LOAD_TARGET_MONO'   : _FT_LOAD_TARGET_(FT_RENDER_MODE_MONO),
    'FT_LOAD_TARGET_LCD'    : _FT_LOAD_TARGET_(FT_RENDER_MODE_LCD),
    'FT_LOAD_TARGET_LCD_V'  : _FT_LOAD_TARGET_(FT_RENDER_MODE_LCD_V) }
globals().update(FT_LOAD_TARGETS)
#def FT_LOAD_TARGET_MODE(x):
#    return (x >> 16) & 15



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  pycairo/cairocffi-based FreeType example - Copyright 2017 Hin-Tak Leung
#  Distributed under the terms of the new BSD license.
#
#  rewrite of the numply,matplotlib-based example from Nicolas P. Rougier
#
# -----------------------------------------------------------------------------
#
# Direct translation of example 1 from the freetype tutorial:
# http://www.freetype.org/freetype2/docs/tutorial/step1.html
#
# Except we uses FreeType's own trigonometric functions instead of those
# from the system/python's math library.


from cairo import Context, ImageSurface, FORMAT_A8
from bitmap_to_surface import make_image_surface

from freetype.raw import *
from PIL import Image

WIDTH, HEIGHT = 640, 480
image = ImageSurface(FORMAT_A8, WIDTH, HEIGHT)
ctx = Context(image)

def to_c_str(text):
    ''' Convert python strings to null terminated c strings. '''
    cStr = create_string_buffer(text.encode(encoding='UTF-8'))
    return cast(pointer(cStr), POINTER(c_char))

def draw_bitmap( bitmap, x, y):
    global image, ctx
    # cairo does not like zero-width surface
    if (bitmap.width > 0):
        glyph_surface = make_image_surface(bitmap)
        ctx.set_source_surface(glyph_surface, x, y)
        ctx.paint()

def main():

    library = FT_Library()
    matrix  = FT_Matrix()
    face    = FT_Face()
    pen     = FT_Vector()
    filename= 'Vera.ttf'
    text    = 'Hello World !'
    num_chars = len(text)
    # FT_Angle is a 16.16 fixed-point value expressed in degrees.
    angle   = FT_Angle(25 * 65536)

    # initialize library, error handling omitted
    error = FT_Init_FreeType( byref(library) )

    # create face object, error handling omitted
    error = FT_New_Face( library, to_c_str(filename), 0, byref(face) )


    # set character size: 50pt at 100dpi, error handling omitted
    error = FT_Set_Char_Size( face, 50 * 64, 0, 100, 0 )
    slot = face.contents.glyph

    # set up matrix
    matrix.xx = FT_Cos( angle )
    matrix.xy = - FT_Sin( angle )
    matrix.yx = FT_Sin( angle )
    matrix.yy = FT_Cos( angle )

    # the pen position in 26.6 cartesian space coordinates; */
    # start at (300,200) relative to the upper left corner  */
    pen.x = 200 * 64;
    pen.y = ( HEIGHT - 300 ) * 64

    for n in range(num_chars):
        # set transformation
        FT_Set_Transform( face, byref(matrix), byref(pen) )

        # load glyph image into the slot (erase previous one)
        charcode = ord(text[n])
        index = FT_Get_Char_Index( face, charcode )
        FT_Load_Glyph( face, index, FT_LOAD_RENDER )

        # now, draw to our target surface (convert position)
        draw_bitmap( slot.contents.bitmap,
                     slot.contents.bitmap_left,
                     HEIGHT - slot.contents.bitmap_top )

        # increment pen position
        pen.x += slot.contents.advance.x
        pen.y += slot.contents.advance.y

    FT_Done_Face(face)
    FT_Done_FreeType(library)

    image.flush()
    image.write_to_png("example_1-cairo.png")
    Image.open("example_1-cairo.png").show()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
'''
Glyph bitmap monochrome rendering
'''
from freetype import *

if __name__ == '__main__':
    import numpy
    import matplotlib.pyplot as plt

    face = Face('./Vera.ttf')
    face.set_char_size( 48*64 )
    face.load_char('S', FT_LOAD_RENDER )
    bitmap = face.glyph.bitmap
    width  = face.glyph.bitmap.width
    rows   = face.glyph.bitmap.rows
    pitch  = face.glyph.bitmap.pitch

    data = []
    for i in range(rows):
        data.extend(bitmap.buffer[i*pitch:i*pitch+width])
    Z = numpy.array(data,dtype=numpy.ubyte).reshape(rows, width)
    plt.imshow(Z, interpolation='nearest', cmap=plt.cm.gray, origin='lower')
    plt.show()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2017 Hin-Tak Leung
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
"""
An enumeration type that lists the curve tags supported by FreeType 2.

Each point of an outline has a specific tag which indicates whether it
describes a point used to control a line segment or an arc. The tags can take
the following values:

FT_CURVE_TAG_ON

  Used when the point is ‘on’ the curve. This corresponds to start and end
  points of segments and arcs. The other tags specify what is called an ‘off’
  point, i.e., a point which isn't located on the contour itself, but serves
  as a control point for a Bézier arc.

FT_CURVE_TAG_CONIC

  Used for an ‘off’ point used to control a conic Bézier arc.

FT_CURVE_TAG_CUBIC

  Used for an ‘off’ point used to control a cubic Bézier arc.


FT_Curve_Tag_On, FT_Curve_Tag_Conic, FT_Curve_Tag_Cubic are their
correspondning mixed-case aliases.

Use the FT_CURVE_TAG(tag) macro to filter out other, internally used flags.

"""

FT_CURVE_TAGS = {
    'FT_CURVE_TAG_ON'    : 1,
    'FT_CURVE_TAG_CONIC' : 0,
    'FT_CURVE_TAG_CUBIC' : 2}
globals().update(FT_CURVE_TAGS)

FT_Curve_Tag_On     = FT_CURVE_TAG_ON
FT_Curve_Tag_Conic  = FT_CURVE_TAG_CONIC
FT_Curve_Tag_Cubic  = FT_CURVE_TAG_CUBIC

def FT_CURVE_TAG( flag ):
    return ( flag & 3 )

# FreeType itself does not have mixed-case macros
FT_Curve_Tag = FT_CURVE_TAG



import glob
import os

import freetype
import pytest

test_folder = os.path.realpath(os.path.dirname(__file__))


def test_load_ft_face():
    """A smoke test."""
    p = os.path.join(test_folder, "..", "examples", "Vera.ttf")
    assert freetype.Face(p)


def test_load_ft_face_from_memory():
    """Another smoke test."""
    p = os.path.join(test_folder, "..", "examples", "Vera.ttf")
    with open(p, mode="rb") as f:
        assert freetype.Face(f)

    with open(p, mode="rb") as f:
        byte_stream = f.read()
    assert freetype.Face.from_bytes(byte_stream)


def test_bundle_version():
    module_dir = os.path.dirname(freetype.__file__)
    shared_object = glob.glob(os.path.join(module_dir, "libfreetype*"))
    if shared_object:
        import re
        p = os.path.join(test_folder, "..", "setup-build-freetype.py")
        with open(p) as f:
            m = re.findall(r"freetype-(\d+)\.(\d+)\.?(\d+)?\.tar", f.read())
        version = m[0]
        if not version[2]:
            version = (int(version[0]), int(version[1]), 0)
        else:
            version = (int(version[0]), int(version[1]), int(version[2]))
        assert freetype.version() == version
    else:
        pytest.skip("Not using a bundled FreeType library.")


def test_get_ft_face_properties():
    p = os.path.join(test_folder, "..", "examples", "Vera.ttf")
    f = freetype.Face(p)
    assert f.family_name == b"Bitstream Vera Sans"
    assert f.get_format() == b"TrueType"
    assert f.get_glyph_name(20) == b"one"
    assert f.get_fstype() == ("FT_FSTYPE_PREVIEW_AND_PRINT_EMBEDDING", 4)
    assert f.get_first_char() == (32, 3)
    assert f.is_sfnt
    assert f.ascender == 1901
    assert f.descender == -483
    assert f.num_glyphs == 268
    assert f.style_name == b"Roman"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
'''
Glyph outline rendering
'''
from freetype import *

if __name__ == '__main__':
    import numpy
    import matplotlib.pyplot as plt

    face = Face('./Vera.ttf')
    face.set_char_size( 4*48*64 )
    flags = FT_LOAD_DEFAULT | FT_LOAD_NO_BITMAP
    face.load_char('S', flags )
    slot = face.glyph
    glyph = slot.get_glyph()
    stroker = Stroker( )
    stroker.set(64, FT_STROKER_LINECAP_ROUND, FT_STROKER_LINEJOIN_ROUND, 0 )
    glyph.stroke( stroker , True )
    blyph = glyph.to_bitmap(FT_RENDER_MODE_NORMAL, Vector(0,0), True )
    bitmap = blyph.bitmap
    width, rows, pitch = bitmap.width, bitmap.rows, bitmap.pitch
    top, left = blyph.top, blyph.left
    data = []
    for i in range(rows):
        data.extend(bitmap.buffer[i*pitch:i*pitch+width])
    Z = numpy.array(data,dtype=numpy.ubyte).reshape(rows, width)
    plt.figure(figsize=(6,8))
    plt.imshow(Z, interpolation='nearest', cmap=plt.cm.gray_r, origin='lower')
    plt.show()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
"""
A list of valid values for the 'encoding_id' for TT_PLATFORM_APPLE_UNICODE
charmaps and name entries.


TT_APPLE_ID_DEFAULT

  Unicode version 1.0.


TT_APPLE_ID_UNICODE_1_1

  Unicode 1.1; specifies Hangul characters starting at U+34xx.


TT_APPLE_ID_ISO_10646

  Deprecated (identical to preceding).


TT_APPLE_ID_UNICODE_2_0

  Unicode 2.0 and beyond (UTF-16 BMP only).


TT_APPLE_ID_UNICODE_32

  Unicode 3.1 and beyond, using UTF-32.


TT_APPLE_ID_VARIANT_SELECTOR

  From Adobe, not Apple. Not a normal cmap. Specifies variations on a real
  cmap.
"""
TT_APPLE_IDS = {
    'TT_APPLE_ID_DEFAULT'          : 0,
    'TT_APPLE_ID_UNICODE_1_1'      : 1,
    'TT_APPLE_ID_ISO_10646'        : 2,
    'TT_APPLE_ID_UNICODE_2_0'      : 3,
    'TT_APPLE_ID_UNICODE_32'       : 4,
    'TT_APPLE_ID_VARIANT_SELECTOR' : 5 }
globals().update(TT_APPLE_IDS)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
"""
A list of bit flags that inform client applications of embedding and
subsetting restrictions associated with a font.

FT_FSTYPE_INSTALLABLE_EMBEDDING

  Fonts with no fsType bit set may be embedded and permanently installed on
  the remote system by an application.


FT_FSTYPE_RESTRICTED_LICENSE_EMBEDDING

  Fonts that have only this bit set must not be modified, embedded or exchanged
  in any manner without first obtaining permission of the font software
  copyright owner.


FT_FSTYPE_PREVIEW_AND_PRINT_EMBEDDING

  If this bit is set, the font may be embedded and temporarily loaded on the
  remote system. Documents containing Preview & Print fonts must be opened
  'read-only'; no edits can be applied to the document.


FT_FSTYPE_EDITABLE_EMBEDDING

  If this bit is set, the font may be embedded but must only be installed
  temporarily on other systems. In contrast to Preview & Print fonts,
  documents containing editable fonts may be opened for reading, editing is
  permitted, and changes may be saved.


FT_FSTYPE_NO_SUBSETTING

  If this bit is set, the font may not be subsetted prior to embedding.


FT_FSTYPE_BITMAP_EMBEDDING_ONLY

  If this bit is set, only bitmaps contained in the font may be embedded; no
  outline data may be embedded. If there are no bitmaps available in the font,
  then the font is unembeddable.
"""

FT_FSTYPES = {'FT_FSTYPE_INSTALLABLE_EMBEDDING'        : 0x0000,
              'FT_FSTYPE_RESTRICTED_LICENSE_EMBEDDING' : 0x0002,
              'FT_FSTYPE_PREVIEW_AND_PRINT_EMBEDDING'  : 0x0004,
              'FT_FSTYPE_EDITABLE_EMBEDDING'           : 0x0008,
              'FT_FSTYPE_NO_SUBSETTING'                : 0x0100,
              'FT_FSTYPE_BITMAP_EMBEDDING_ONLY'        : 0x0200,}
globals().update(FT_FSTYPES)
ft_fstype_installable_embedding  = FT_FSTYPE_INSTALLABLE_EMBEDDING
ft_fstype_restricted_license_embedding = FT_FSTYPE_RESTRICTED_LICENSE_EMBEDDING
ft_fstype_preview_and_print_embedding = FT_FSTYPE_PREVIEW_AND_PRINT_EMBEDDING
ft_fstype_editable_embedding = FT_FSTYPE_EDITABLE_EMBEDDING
ft_fstype_no_subsetting = FT_FSTYPE_NO_SUBSETTING
ft_fstype_bitmap_embedding_only = FT_FSTYPE_BITMAP_EMBEDDING_ONLY



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier
#  Distributed under the terms of the new BSD license.
#
# -----------------------------------------------------------------------------
"""
A list of bit-field constants used with FT_Load_Glyph to indicate what kind
of operations to perform during glyph loading.


FT_LOAD_DEFAULT

  Corresponding to 0, this value is used as the default glyph load
  operation. In this case, the following happens:

  1. FreeType looks for a bitmap for the glyph corresponding to the face's
     current size. If one is found, the function returns. The bitmap data can
     be accessed from the glyph slot (see note below).

  2. If no embedded bitmap is searched or found, FreeType looks for a scalable
     outline. If one is found, it is loaded from the font file, scaled to
     device pixels, then 'hinted' to the pixel grid in order to optimize
     it. The outline data can be accessed from the glyph slot (see note below).

  Note that by default, the glyph loader doesn't render outlines into
  bitmaps. The following flags are used to modify this default behaviour to
  more specific and useful cases.


FT_LOAD_NO_SCALE

  Don't scale the outline glyph loaded, but keep it in font units.

  This flag implies FT_LOAD_NO_HINTING and FT_LOAD_NO_BITMAP, and unsets
  FT_LOAD_RENDER.


FT_LOAD_NO_HINTING

  Disable hinting. This generally generates 'blurrier' bitmap glyph when the
  glyph is rendered in any of the anti-aliased modes. See also the note below.

  This flag is implied by FT_LOAD_NO_SCALE.


FT_LOAD_RENDER

  Call FT_Render_Glyph after the glyph is loaded. By default, the glyph is
  rendered in FT_RENDER_MODE_NORMAL mode. This can be overridden by
  FT_LOAD_TARGET_XXX or FT_LOAD_MONOCHROME.

  This flag is unset by FT_LOAD_NO_SCALE.


FT_LOAD_NO_BITMAP

  Ignore bitmap strikes when loading. Bitmap-only fonts ignore this flag.

  FT_LOAD_NO_SCALE always sets this flag.


FT_LOAD_VERTICAL_LAYOUT

  Load the glyph for vertical text layout. Don't use it as it is problematic
  currently.


FT_LOAD_FORCE_AUTOHINT

  Indicates that the auto-hinter is preferred over the font's native
  hinter. See also the note below.


FT_LOAD_CROP_BITMAP

  Indicates that the font driver should crop the loaded bitmap glyph (i.e.,
  remove all space around its black bits). Not all drivers implement this.


FT_LOAD_PEDANTIC

  Indicates that the font driver should perform pedantic verifications during
  glyph loading. This is mostly used to detect broken glyphs in fonts. By
  default, FreeType tries to handle broken fonts also.


FT_LOAD_IGNORE_GLOBAL_ADVANCE_WIDTH

  Indicates that the font driver should ignore the global advance width defined
  in the font. By default, that value is used as the advance width for all
  glyphs when the face has FT_FACE_FLAG_FIXED_WIDTH set.

  This flag exists for historical reasons (to support buggy CJK fonts).


FT_LOAD_NO_RECURSE

  This flag is only used internally. It merely indicates that the font driver
  should not load composite glyphs recursively. Instead, it should set the
  'num_subglyph' and 'subglyphs' values of the glyph slot accordingly, and set
  'glyph->format' to FT_GLYPH_FORMAT_COMPOSITE.

  The description of sub-glyphs is not available to client applications for now.

  This flag implies FT_LOAD_NO_SCALE and FT_LOAD_IGNORE_TRANSFORM.


FT_LOAD_IGNORE_TRANSFORM

  Indicates that the transform matrix set by FT_Set_Transform should be ignored.


FT_LOAD_MONOCHROME

  This flag is used with FT_LOAD_RENDER to indicate that you want to render an
  outline glyph to a 1-bit monochrome bitmap glyph, with 8 pixels packed into
  each byte of the bitmap data.

  Note that this has no effect on the hinting algorithm used. You should rather
  use FT_LOAD_TARGET_MONO so that the monochrome-optimized hinting algorithm is
  used.


FT_LOAD_LINEAR_DESIGN

  Indicates that the 'linearHoriAdvance' and 'linearVertAdvance' fields of
  FT_GlyphSlotRec should be kept in font units. See FT_GlyphSlotRec for
  details.


FT_LOAD_NO_AUTOHINT

  Disable auto-hinter. See also the note below.


FT_LOAD_COLOR

  This flag is used to request loading of color embedded-bitmap images. The
  resulting color bitmaps, if available, will have the FT_PIXEL_MODE_BGRA
  format. When the flag is not used and color bitmaps are found, they will be
  converted to 256-level gray bitmaps transparently. Those bitmaps will be in
  the FT_PIXEL_MODE_GRAY format.


FT_LOAD_NO_SVG

   [Since 2.13.1] Ignore SVG glyph data when loading.


FT_LOAD_COMPUTE_METRICS
  Compute glyph metrics from the glyph data, without the use of bundled metrics
  tables (for example, the `hdmx' table in TrueType fonts).  This flag is
  mainly used by font validating or font editing applications, which need to
  ignore, verify, or edit those tables.

  Currently, this flag is only implemented for TrueType fonts.


FT_LOAD_BITMAP_METRICS_ONLY
  Request loading of the metrics and bitmap image information of a (possibly
  embedded) bitmap glyph without allocating or copying the bitmap image data
  itself.  No effect if the target glyph is not a bitmap image.

  This flag unsets FT_LOAD_RENDER.
"""

FT_LOAD_FLAGS = { 'FT_LOAD_DEFAULT'                      : 0x0,
                  'FT_LOAD_NO_SCALE'                     : 0x1,
                  'FT_LOAD_NO_HINTING'                   : 0x2,
                  'FT_LOAD_RENDER'                       : 0x4,
                  'FT_LOAD_NO_BITMAP'                    : 0x8,
                  'FT_LOAD_VERTICAL_LAYOUT'              : 0x10,
                  'FT_LOAD_FORCE_AUTOHINT'               : 0x20,
                  'FT_LOAD_CROP_BITMAP'                  : 0x40,
                  'FT_LOAD_PEDANTIC'                     : 0x80,
                  'FT_LOAD_ADVANCE_ONLY'                 : 0x100,
                  'FT_LOAD_IGNORE_GLOBAL_ADVANCE_WIDTH'  : 0x200,
                  'FT_LOAD_NO_RECURSE'                   : 0x400,
                  'FT_LOAD_IGNORE_TRANSFORM'             : 0x800,
                  'FT_LOAD_MONOCHROME'                   : 0x1000,
                  'FT_LOAD_LINEAR_DESIGN'                : 0x2000,
                  'FT_LOAD_SBITS_ONLY'                   : 0x4000,
                  'FT_LOAD_NO_AUTOHINT'                  : 0x8000,
                  'FT_LOAD_COLOR'                        : 0x100000,
                  'FT_LOAD_COMPUTE_METRICS'              : 0x200000,
                  'FT_LOAD_BITMAP_METRICS_ONLY'          : 0x400000,
                  'FT_LOAD_NO_SVG'                       : 0x800000 }
globals().update(FT_LOAD_FLAGS)
